import os
import json
import copy
import inspect
import warnings
from typing import Union
from collections import OrderedDict

import sklearn
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from easy_mpl import parallel_coordinates, hist
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn.model_selection import ParameterGrid, ParameterSampler

try:
    import plotly
except ImportError:
    plotly = None

try:
    import skopt
    from skopt import gp_minimize
    from skopt import BayesSearchCV
    from skopt.space.space import Space
    from skopt.utils import use_named_args
    from skopt.space.space import Dimension
    from skopt.plots import plot_convergence, plot_evaluations
except ImportError:
    skopt, gp_minimize, BayesSearchCV, Space, _Real, use_named_args = None, None, None, None, None, None
    Dimension, _Integer, _Categorical, plot_evaluations, plot_convergence = None, None, None, None, None

if skopt is not None:
    from skopt import forest_minimize

try:
    import hyperopt
    from hyperopt.pyll.base import Apply
    from hyperopt import fmin as fmin_hyperopt
    from hyperopt import tpe, STATUS_OK, Trials, rand
except ImportError:
    hyperopt, fmin_hyperopt, tpe, atpe, Trials, rand, Apply = None, None, None, None, None, None, None
    space_eval, miscs_to_idxs_vals = None, None

try:  # atpe is only available in later versions of hyperopt
    from hyperopt import atpe
except ImportError:
    atpe = None

try:
    import optuna
    from optuna.study import Study
    from optuna.visualization import plot_contour
except ImportError:
    optuna, plot_contour = None, None
    Study = None

from ai4water.utils.utils import JsonEncoder
from .utils import plot_convergences
from .utils import get_one_tpe_x_iter
from ai4water.utils.utils import Jsonize, dateandtime_now
from .utils import to_skopt_as_dict
from .utils import post_process_skopt_results
from .utils import to_skopt_space
from .utils import save_skopt_results
from ._space import Categorical, Real, Integer
from .utils import sort_x_iters, x_iter_for_tpe
from .utils import loss_histogram, plot_hyperparameters, plot_edf


try:
    from.testing import plot_param_importances
except ModuleNotFoundError:
    plot_param_importances = None

# TODO RayTune libraries under the hood https://docs.ray.io/en/master/tune/api_docs/suggestion.html#summary
# TODO add generic algorithm, deap/pygad
# TODO skopt provides functions other than gp_minimize, see if they are useful and can be used.
# todo loading gpmin_results is not consistent.

SEP = os.sep
COUNTER = 0

ALGORITHMS = {
    'bayes': {},
    'bayes_rf': {'name': 'decision_tree', 'backend': ['skopt']},
    'gbrt': {'name': 'gradient-boosted-tree regression', 'backend': ['skopt']},
    'tpe': {'name': 'Tree of Parzen Estimators', 'backend': ['hyperopt', 'optuna']},
    'atpe': {'name': 'Adaptive Tree of Parzen Estimators', 'backend': ['hyperopt']},
    'random': {'name': 'random search', 'backend': ['sklearn', 'optuna', 'hyperopt']},
    'grid': {'name': 'grid search', 'backend': ['sklearn', 'optuna']},
    'cmaes': {'name': 'Covariance Matrix Adaptation Evolution Strategy', 'backend': ['optuna']}
}


class HyperOpt(object):
    """
    The purpose of this class is to provide a uniform and simplifed interface to
    use `hyperopt`, `optuna`, `scikit-optimize` and `scikit-learn` based hyperparameter
    optimization methods. Ideally this class should provide all the functionalities of
    beforementioned libaries with a uniform interface. It however also complements
    these libraries by combining their functionalities and adding some additional
    functionalities to them. On the other hand this class should not limit or
    complicate the use of its underlying libraries. This means all the functionalities
    of underlying libraries are available in this class as well. Moreover, you can
    use this class just as you use one of its underlying library.

    The purpose here is to make a class which allows application of any of the
    available optimization methods on any type of model/classifier/regressor. If the
    classifier/regressor is of sklearn-based, then for random search, we use
    RanddomSearchCV_, for grid search, we use GridSearchCV_ and for Bayesian, we
    use BayesSearchCV_ . On the other hand, if the model is not sklearn-based, you
    will still be able to implement any of the three methods. In such case, the
    bayesian_ will be implemented using `gp_minimize`. Random search and grid search
    will be done by simple iterating over the sample space generated as in sklearn
    based samplers. However, the post-processing of the results is (supposed to be)
    done same as is done in RandomSearchCV and GridSearchCV.

    The class is expected to pass all the tests written in sklearn or skopt for
    corresponding classes.

    For detailed use of this class see this `hpo_tutorial`_

    Attributes
    --------------
    - results dict:
    - gpmin_results dict:
    - skopt_results :
    - hp_space :
    - space
    - skopt_space :
    - space dict:
    - title str: name of the folder in which all results will be saved. By
        default this is same as name of `algorithm`. For `AI4Water` based
            models, this is more detailed, containing problem type etc.


    Methods
    -------
    - eval_with_best: evaluates the objective_fn on best parameters
    - best_paras(): returns the best parameters from optimization.

    The following examples illustrate how we can uniformly apply different optimization algorithms.

    Examples
    --------
        >>> from ai4water import Model
        >>> from ai4water.hyperopt import HyperOpt, Categorical, Integer, Real
        >>> from ai4water.datasets import busan_beach
        >>> from SeqMetrics import RegressionMetrics
        >>> data = busan_beach()
        >>> input_features = ['tide_cm', 'wat_temp_c', 'sal_psu', 'air_temp_c', 'pcp_mm', 'pcp3_mm']
        >>> output_features = ['tetx_coppml']

        We have to define an objective function which will take keyword arguments
        and return a scaler value as output. This scaler value will be minized during optimzation

        >>> def objective_fn(**suggestion)->float:
        ...   # the objective function must receive new parameters as keyword arguments
        ...    model = Model(
        ...        input_features=input_features,
        ...        output_features=output_features,
        ...        model={"XGBRegressor": suggestion},
        ...        train_data='random',
        ...        verbosity=0)
        ...
        ...    model.fit(data=data)
        ...
        ...    t, p = model.predict(return_true=True)
        ...    mse = RegressionMetrics(t, p).mse()
        ...    # the objective function must return a scaler value which needs to be minimized
        ...    return mse

        Define search space
        The search splace determines pool from which parameters are chosen during optimization.

        >>> num_samples=5   # only relavent for random and grid search
        >>> search_space = [
        ...    Categorical(['gbtree', 'dart'], name='booster'),
        ...    Integer(low=1000, high=2000, name='n_estimators', num_samples=num_samples),
        ...    Real(low=1.0e-5, high=0.1, name='learning_rate', num_samples=num_samples)
        ...]
        ...
        ... # Using Baysian with gaussian processes
        >>> optimizer = HyperOpt('bayes', objective_fn=objective_fn, param_space=search_space,
        ...                     num_iterations=num_iterations )
        >>> optimizer.fit()

        Using TPE with optuna

        >>> num_iterations = 10
        >>> optimizer = HyperOpt('tpe', objective_fn=objective_fn, param_space=search_space,
        ...                     backend='optuna',
        ...                     num_iterations=num_iterations )
        >>> optimizer.fit()

        Using cmaes with optuna

        >>> optimizer = HyperOpt('cmaes', objective_fn=objective_fn, param_space=search_space,
        ...                     backend='optuna',
        ...                     num_iterations=num_iterations )
        >>> optimizer.fit()

        Using random with optuna, we can also try hyperopt and sklearn as backend for random algorithm

        >>> optimizer = HyperOpt('random', objective_fn=objective_fn, param_space=search_space,
        ...                     backend='optuna',
        ...                     num_iterations=num_iterations )
        >>> optimizer.fit()

        Using TPE of hyperopt

        >>> optimizer = HyperOpt('tpe', objective_fn=objective_fn, param_space=search_space,
        ...                     backend='hyperopt',
        ...                     num_iterations=num_iterations )
        >>> optimizer.fit()

        Using grid with sklearn

        >>> optimizer = HyperOpt('grid', objective_fn=objective_fn, param_space=search_space,
        ...                     backend='sklearn',
        ...                     num_iterations=num_iterations )
        >>> optimizer.fit()

        Backward compatability
        The following shows some tweaks with hyperopt to make its working
        compatible with its underlying libraries.
        using grid search with AI4Water

        using Bayesian with custom objective_fn

        >>> def f(x, noise_level=0.1):
        ...      return np.sin(5 * x[0]) * (1 - np.tanh(x[0] ** 2)) + np.random.randn() * noise_level
        ...
        >>> opt = HyperOpt("bayes",
        ...           objective_fn=f,
        ...           param_space=[Categorical([32, 64, 128, 256], name='lstm_units'),
        ...                        Categorical(categories=["relu", "elu", "leakyrelu"], name="dense_actfn")
        ...                        ],
        ...           acq_func='EI',  # Expected Improvement.
        ...           n_calls=50,     #number of iterations
        ...           x0=[32, "relu"],  # inital value of optimizing parameters
        ...           n_random_starts=3,  # the number of random initialization points
        ...           )
        >>> opt_results = opt.fit()

        using Bayesian with custom objective_fn and named args

        >>> def f(noise_level=0.1, **kwargs):
        ...    x = kwargs['x']
        ...    return np.sin(5 * x[0]) * (1 - np.tanh(x[0] ** 2)) + np.random.randn() * noise_level
        >>> opt = HyperOpt("bayes",
        ...           objective_fn=f,
        ...           param_space=[Categorical([32, 64, 128, 256], name='lstm_units'),
        ...                        Categorical(categories=["relu", "elu", "leakyrelu"], name="dense_actfn")
        ...                        ],
        ...           acq_func='EI',  # Expected Improvement.
        ...           n_calls=50,     #number of iterations
        ...           x0=[32, "relu"],  # inital value of optimizing parameters
        ...           n_random_starts=3,  # the number of random initialization points
        ...           random_state=2
        ...           )
        >>> opt_results = opt.fit()

    .. _hpo_tutorial:
        https://github.com/AtrCheema/AI4Water/blob/master/examples/hyper_para_opt.ipynb

    .. _GridSearchCV:
        https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.GridSearchCV.html

    .. _RanddomSearchCV:
        https://scikit-learn.org/stable/modules/generated/sklearn.model_selection.RandomizedSearchCV.html

    .. _BayesSearchCV:
        https://scikit-optimize.github.io/stable/modules/generated/skopt.BayesSearchCV.html

    .. _bayesian:
        https://github.com/scikit-optimize/scikit-optimize/blob/9334d50a1ad5c9f7c013a1c1cb95313a54b83168/examples/bayesian-optimization.py#L109

    """

    def __init__(
            self,
            algorithm: str, *,
            param_space,
            objective_fn=None,
            eval_on_best: bool = False,
            backend: str = None,
            opt_path: str = None,
            process_results: bool = True,
            verbosity: int = 1,
            **kwargs
    ):

        """
        Initializes the class

        Parameters
        ---------
            algorithm str:
                must be one of "random", "grid" "bayes", "bayes_rf", and "tpe", defining which
                optimization algorithm to use.
            objective_fn callable:
                Any callable function whose returned value is to be minimized.
                It can also be either sklearn/xgboost based regressor/classifier.
            backend str:
                Defines which backend library to use for the `algorithm`. For
                example the user can specify whether to use `optuna` or `hyper_opt`
                or `sklearn` for `grid` algorithm.
            param_space list/dict:
                the search space of parameters to be optimized. We recommend the use
                of Real, Integer and categorical classes from [ai4water.hyperopt][ai4water.hyperopt.Integer]
                (not from skopt.space). These classes allow a uniform way of defining
                the parameter space for all the underlying libraries. However, to
                make this class work exactly similar to its underlying libraries,
                the user can also define parameter space as is defined in its
                underlying libraries. For example, for hyperopt based method like
                'tpe' the parameter space can be specified as in the examples of
                hyperopt library. In case the code breaks, please report.
            eval_on_best bool:
                if True, then after optimization, the objective_fn will
                be evaluated on best parameters and the results will be stored in the
                folder named "best" inside `title` folder.
            opt_path : path to save the results
            verbosity : determines amount of information being printed
            kwargs dict:
                Any additional keyword arguments will for the underlying optimization
                algorithm. In case of using AI4Water model, these must be arguments
                which are passed to AI4Water's Model class.
        """
        if algorithm not in ALGORITHMS:
            raise ValueError(f"""Invalid value of algorithm provided. Allowd values for algorithm"
                                are {list(ALGORITHMS.keys())}. 
                                You provided {algorithm}""")

        self.objective_fn = objective_fn
        self.algorithm = algorithm
        self.backend = backend
        self.param_space = param_space
        self.original_space = param_space       # todo self.space and self.param_space should be combined.
        self.title = self.algorithm
        self.results = {}  # internally stored results
        self.gpmin_results = None  #
        self.data = None
        self.eval_on_best = eval_on_best
        self.opt_path = opt_path
        self.process_results = process_results
        self.objective_fn_is_dl = False
        self.verbosity = verbosity

        self.gpmin_args = self.check_args(**kwargs)

        if self.use_sklearn:
            if self.algorithm == "random":
                self.optfn = RandomizedSearchCV(estimator=objective_fn, param_distributions=param_space, **kwargs)
            else:
                self.optfn = GridSearchCV(estimator=objective_fn, param_grid=param_space, **kwargs)

        elif self.use_skopt_bayes:
            self.optfn = BayesSearchCV(estimator=objective_fn, search_spaces=param_space, **kwargs)

    @property
    def backend(self):
        return self._backend

    @backend.setter
    def backend(self, x):
        if x is not None:
            assert x in ['optuna', 'hyperopt', 'sklearn', 'skopt'], f"""
Backend must be one of hyperopt, optuna or sklearn but is is {x}"""
        if self.algorithm == 'tpe':
            if x is None:
                x = 'optuna'
            assert x in ['optuna', 'hyperopt']
        elif self.algorithm == 'cmaes':
            if x is None:
                x = 'optuna'
            assert x == 'optuna'
        elif self.algorithm == 'atpe':
            if x is None:
                x = 'hyperopt'
            assert x == 'hyperopt'
        elif self.algorithm == 'random':
            if x is None:
                x = 'sklearn'
            assert x in ['optuna', 'hyperopt', 'sklearn']
        elif self.algorithm == 'grid':
            if x is None:
                x = 'sklearn'
            assert x in ['sklearn', 'optuna']
        elif self.algorithm in ['bayes', "bayes_rf"]:
            if x is None:
                x = 'skopt'
        else:
            raise ValueError
        if x == 'hyperopt' and hyperopt is None:
            raise ValueError(f"You must install `hyperopt` to use it as backend for {self.algorithm} algorithm.")
        if x == 'optuna' and optuna is None:
            raise ValueError(f"You must install optuna to use `optuna` as backend for {self.algorithm} algorithm")
        self._backend = x

    @property
    def title(self):
        return self._title

    @title.setter
    def title(self, x):
        self._title = x + '_' + str(dateandtime_now())

    @property
    def objective_fn_is_dl(self):
        return self._objective_fn_is_dl

    @objective_fn_is_dl.setter
    def objective_fn_is_dl(self, x):
        self._objective_fn_is_dl = x

    def check_args(self, **kwargs):
        kwargs = copy.deepcopy(kwargs)

        if 'n_initial_points' in kwargs:
            if int(''.join(skopt.__version__.split('.')[1])) < 8:
                raise ValueError(f"""
                        'n_initial_points' argument is not available in skopt version < 0.8.
                        However you are using skopt version {skopt.__version__} .
                        See https://scikit-optimize.github.io/stable/modules/generated/skopt.gp_minimize.html#skopt.gp_minimize
                        for more details.
                        """"")
        if 'x0' in kwargs and self.algorithm in ['tpe', 'atpe', 'random', 'grid', 'cmaes']:
            kwargs.pop('x0')
        return kwargs

    def __getattr__(self, item):
        # TODO, not sure if this is the best way but venturing since it is done by the legend
        #  here https://github.com/philipperemy/n-beats/blob/master/nbeats_keras/model.py#L166
        # Since it was not possible to inherit this class from BaseSearchCV and BayesSearchCV at the same time, this
        # hack makes sure that all the functionalities of GridSearchCV, RandomizeSearchCV and BayesSearchCV are also
        # available with class.
        if self.use_sklearn or self.use_skopt_bayes:
            return getattr(self.optfn, item)
        else:
            raise AttributeError(f"HyperOpt does not have attribute {item}")

    @property
    def param_space(self):
        return self._param_space

    @param_space.setter
    def param_space(self, x):
        if self.algorithm in ["bayes", "bayes_rf"]:
            if isinstance(x, dict):
                _param_space = []
                for k, v in x.items():
                    assert isinstance(v, Dimension), f"""
                            space for parameter {k} is of invalid type {v.__class__.__name__}.
                            For {self.algorithm}, it must be of type {Dimension.__name__}
                            """
                    _param_space.append(v)
            else:
                assert isinstance(x, list), f"""
                        param space must be list of parameters but it is of type
                        {x.__class__.__name__}"""
                for space in x:
                    # each element in the list can be a tuple of lower and and upper bounds
                    if not isinstance(space, tuple):
                        assert isinstance(space, Dimension), f"""
                                param space must be one of Integer, Real or Categorical
                                but it is of type {space.__class__.__name__}"""
                _param_space = x

        elif self.algorithm in ["random", "grid"] and self.backend != 'optuna':
            # todo, do we also need to provide grid of sample space for random??
            if isinstance(x, dict):
                _param_space = x
            elif isinstance(x, list):
                _param_space = {}
                for _space in x:
                    assert isinstance(_space, Dimension)
                    _param_space[_space.name] = _space.grid
            else:
                raise ValueError
        elif self.algorithm in ['tpe', 'atpe', 'random'] and self.backend == 'hyperopt':
            if isinstance(x, list):
                # space is provided as list. Either all of them must be hp.space or Dimension.
                if isinstance(x[0], Dimension):
                    _param_space = {}
                    for space in x:
                        assert isinstance(space, Dimension)
                        _param_space[space.name] = space.as_hp()
                elif isinstance(x[0], Apply):
                    _param_space = []
                    for space in x:
                        assert isinstance(space, Apply), f"""invalid space type {space.__class__.__name__}"""
                        _param_space.append(space)
                else:
                    raise NotImplementedError

            elif isinstance(x, Dimension):  # for single hyper-parameter optimization ?
                _param_space = x.as_hp()
            else:
                _param_space = x

        elif self.backend == 'optuna':
            if isinstance(x, list):
                _param_space = {}
                for s in x:
                    assert isinstance(s, Dimension)
                    _param_space[s.name] = s
            elif isinstance(x, dict):
                assert all([isinstance(s, Dimension) for s in x.values()])
                _param_space = x
            else:
                raise NotImplementedError(f"unknown type of space {x.__class__.__name__}")
        else:
            raise ValueError

        self._param_space = _param_space

    def skopt_space(self):
        """Tries to make skopt compatible Space object. If unsuccessful, return None"""
        return to_skopt_space(self.original_space)

    def space(self) -> dict:
        """Returns a skopt compatible space but as dictionary"""
        return to_skopt_as_dict(self.algorithm, self.backend, self.original_space)

    @property
    def use_sklearn(self):
        # will return True if we are to use sklearn's GridSearchCV or RandomSearchCV
        if self.algorithm in ["random", "grid"] and "sklearn" in str(type(self.objective_fn)):
            return True
        return False

    @property
    def use_skopt_bayes(self):
        # will return true if we have to use skopt based BayesSearchCV
        if self.algorithm in ["bayes", "bayes_rf"] and "sklearn" in str(type(self.objective_fn)):
            assert not self.use_sklearn
            return True
        return False

    @property
    def use_skopt_gpmin(self):
        # will return True if we have to use skopt based gp_minimize function. This is to implement Bayesian on
        # non-sklearn based models
        if self.algorithm in ["bayes", "bayes_rf"] and "sklearn" not in str(type(self.objective_fn)):
            assert not self.use_sklearn
            assert not self.use_skopt_bayes
            return True
        return False

    @property
    def use_tpe(self):
        if self.algorithm in ['tpe', 'atpe', 'random'] and self.backend == 'hyperopt':
            return True
        else:
            return False

    @property
    def use_own(self):
        # return True, we have to build our own optimization method.
        if not self.use_sklearn and not self.use_skopt_bayes and not self.use_skopt_gpmin:
            return True
        return False

    @property
    def random_state(self):
        if "random_state" not in self.gpmin_args:
            return np.random.RandomState(313)
        else:
            return np.random.RandomState(self.gpmin_args['random_state'])

    @property
    def num_iterations(self):
        if self.backend == 'sklearn' and self.algorithm == 'grid':
            self.gpmin_args['num_iterations'] = len(ParameterGrid(self.param_space))
        if 'num_iterations' in self.gpmin_args:
            return self.gpmin_args['num_iterations']
        if self.algorithm in ['tpe', 'atpe', 'random'] and self.backend == 'hyperopt':
            return self.gpmin_args.get('max_evals', 9223372036854775807)
        if self.backend == 'optuna':
            return self.gpmin_args.get('n_trials', None)  # default value of n_trials is None in study.optimize()
        if 'n_calls' in self.gpmin_args:
            return self.gpmin_args['n_calls']
        return self.gpmin_args['n_iter']

    @property
    def use_named_args(self):

        argspec = inspect.getfullargspec(self.objective_fn)
        if argspec.varkw is None:
            return False
        elif isinstance(argspec.varkw, str):
            return True
        else:
            raise NotImplementedError

    @property
    def opt_path(self):
        return self._opt_path

    @opt_path.setter
    def opt_path(self, path):
        if path is None:
            path = os.path.join(os.getcwd(), f"results{SEP}" + self.title)
            if not os.path.exists(path):
                os.makedirs(path)
        elif not os.path.exists(path):
            os.makedirs(path)

        self._opt_path = path

    def best_paras(self, as_list=False) -> Union[list, dict]:
        # returns best parameters either as dictionary or as list
        if self.use_skopt_gpmin:
            d = self.xy_of_iterations()
            k = list(dict(sorted(d.items())).keys())[0]
            paras = d[k]
        elif self.backend == 'hyperopt':
            d = get_one_tpe_x_iter(self.trials.best_trial['misc']['vals'], self.hp_space())
            if as_list:
                return list(d.values())
            else:
                return d
        elif self.backend == 'optuna':
            if as_list:
                return list(self.study.best_trial.params.values())
            return self.study.best_trial.params
        elif self.use_skopt_bayes or self.use_sklearn:
            paras = self.optfn.best_params_
        else:
            best_y = list(sorted(self.results.keys()))[0]
            paras = sort_x_iters(self.results[best_y], list(self.param_space.keys()))

        if as_list:
            return list(paras.values())
        return paras

    def fit(self, *args, **kwargs):
        """Makes and calls the underlying fit method"""

        if self.use_sklearn or self.use_skopt_bayes:
            fit_fn = self.optfn.fit

        elif self.use_skopt_gpmin:
            fit_fn = self.own_fit

        elif self.use_own:
            self.predict = self._predict
            if self.algorithm == "grid" and self.backend != 'optuna':
                fit_fn = self.grid_search
            elif self.algorithm == 'random' and self.backend not in ['optuna', 'hyperopt']:
                fit_fn = self.random_search
            elif self.backend == 'hyperopt':
                fit_fn = self.fmin
            elif self.backend == 'optuna':
                fit_fn = self.optuna_objective
            else:
                raise NotImplementedError
        else:
            raise NotImplementedError(f"""No fit function found for algorithm {self.algorithm}
                                          with backend {self.backend}""")
        res = fit_fn(*args, **kwargs)

        serialized = self.serialize()
        fname = os.path.join(self.opt_path, 'serialized.json')
        with open(fname, 'w') as fp:
            json.dump(serialized, fp, sort_keys=True, indent=4, cls=JsonEncoder)

        return res

    def original_para_order(self):
        if isinstance(self.param_space, dict):
            return list(self.param_space.keys())
        elif self.skopt_space() is not None:
            names = []
            for s in self.skopt_space():
                names.append(s.name)
            return names
        else:
            raise NotImplementedError

    def dims(self):
        # this will be used for gp_minimize
        return list(self.param_space)

    def model_for_gpmin(self):
        """
        This function can be called in two cases
            - The user has made its own objective_fn.
            - We make objective_fn using AI4Water and return the error.

        In first case, we just return what user has provided.
        """
        if callable(self.objective_fn) and not self.use_named_args:
            # external function for bayesian but this function does not require named args.
            return self.objective_fn

        dims = self.dims()
        if self.use_named_args:
            # external function and this function accepts named args.
            @use_named_args(dimensions=dims)
            def fitness(**kwargs):
                return self.objective_fn(**kwargs)
            return fitness

        raise ValueError(f"used named args is {self.use_named_args}")

    def own_fit(self):

        if self.algorithm == "bayes":
            minimize_func = gp_minimize
        else: # bayes_rf
            minimize_func = forest_minimize

        kwargs = self.gpmin_args
        if 'num_iterations' in kwargs:
            kwargs['n_calls'] = kwargs.pop('num_iterations')

        try:
            search_result = minimize_func(func=self.model_for_gpmin(),
                                        dimensions=self.dims(),
                                        **kwargs)
        except ValueError as e:
            if int(''.join(sklearn.__version__.split('.')[1])) > 22:
                raise ValueError(f"""
                    For bayesian optimization, If your sklearn version is above 0.23,
                    then this error may be related to 
                    https://github.com/kiudee/bayes-skopt/issues/90 .
                    Try to lower the sklearn version to 0.22 and run again.
                    {e}
                    """)
            else:
                raise ValueError(e)

        # the `space` in search_results may not be in same order as originally provided.
        space = search_result['space']
        if space.__dict__.__len__() > 1:
            ordered_sapce = OrderedDict()
            for k in self.space().keys():
                ordered_sapce[k] = [s for s in space if s.name == k][0]
            search_result['space'] = Space(ordered_sapce.values())

        self.gpmin_results = search_result

        if len(self.results) < 1:
            fv = search_result.func_vals
            xiters = search_result.x_iters
            self.results = {f'{round(k, 8)}_{idx}': self.to_kw(v) for idx, k, v in zip(range(self.num_iterations), fv, xiters)}

        if self.process_results:
            post_process_skopt_results(search_result, self.results, self.opt_path)

            if len(search_result.func_vals)<=100 and self.algorithm != "bayes_rf":
                save_skopt_results(search_result, self.opt_path)

            self._process_results()

        if self.eval_on_best:
            self.eval_with_best()

        return search_result

    def eval_sequence(self, params):

        if self.verbosity > 0:
            print(f"total number of iterations: {len(params)}")
        for idx, para in enumerate(params):

            if self.use_named_args:  # objective_fn is external but uses kwargs
                err = self.objective_fn(**para)
            else:  # objective_fn is external and does not uses keywork arguments
                try:
                    err = self.objective_fn(*list(para.values()))
                except TypeError:
                    raise TypeError(f"""
                        use_named_args argument is set to {self.use_named_args}. If your
                        objective function takes key word arguments, make sure that
                        this argument is set to True during initiatiation of HyperOpt.""")
            err = round(err, 8)

            self.results[err + idx] = sort_x_iters(para, self.original_para_order())

        if self.process_results:
            self._process_results()

        if self.eval_on_best:
            self.eval_with_best()

        return self.results

    def grid_search(self):

        params = list(ParameterGrid(self.param_space))
        self.param_grid = params

        return self.eval_sequence(params)

    def random_search(self):

        for k, v in self.param_space.items():
            if v is None:
                grid = self.space()[k].grid
                assert grid is not None, f"""grid for parameter {k} could not be created. Inferred grid is 
                {grid}. Please either provide the `num_samples` parameter while creating space or explicitly
                provide grid for {k}"""

        param_list = list(ParameterSampler(self.param_space, n_iter=self.num_iterations,
                                           random_state=self.random_state))

        if len(param_list) < self.num_iterations:
            # we need to correct it so that num_iterations gets calculated correctly next time
            self.gpmin_args['n_calls'] = len(param_list)
            self.gpmin_args['n_iter'] = len(param_list)

        self.param_grid = param_list

        return self.eval_sequence(param_list)

    def optuna_objective(self, **kwargs):

        if self.verbosity == 0:
            optuna.logging.set_verbosity(optuna.logging.ERROR)

        sampler = {
            'tpe': optuna.samplers.TPESampler,
            'cmaes': optuna.samplers.CmaEsSampler,
            'random': optuna.samplers.RandomSampler,
            'grid': optuna.samplers.GridSampler
        }

        def objective(trial):
            suggestion = {}
            for space_name, _space in self.param_space.items():
                suggestion[space_name] = _space.suggest(trial)
            return self.objective_fn(**suggestion)

        if self.algorithm in ['tpe', 'cmaes', 'random']:
            study = optuna.create_study(direction='minimize', sampler=sampler[self.algorithm]())
        else:
            space = {s.name: s.grid for s in self.skopt_space()}
            study = optuna.create_study(sampler=sampler[self.algorithm](space))
        study.optimize(objective, n_trials=self.num_iterations)
        setattr(self, 'study', study)

        if self.process_results:
            self._process_results()

        return study

    def fmin(self, **kwargs):

        suggest_options = {
            'tpe': tpe.suggest,
            'random': rand.suggest
        }
        if atpe is not None:
            suggest_options.update({'atpe': atpe.suggest})

        trials = Trials()
        model_kws = self.gpmin_args
        if 'num_iterations' in model_kws:
            model_kws['max_evals'] = model_kws.pop('num_iterations')

        space = self.hp_space()
        if self.use_named_args:
            def objective_fn(kws):
                # the objective function in hyperopt library receives a dictionary
                return self.objective_fn(**kws)
            objective_f = objective_fn
        else:
            objective_f = self.objective_fn

            if len(self.space()) > 1:
                space = list(self.hp_space().values())
            elif len(self.space()) == 1:
                space = list(self.hp_space().values())[0]
            else:
                raise NotImplementedError

        best = fmin_hyperopt(objective_f,
                             space=space,
                             algo=suggest_options[self.algorithm],
                             trials=trials,
                             **kwargs,
                             **model_kws)

        with open(os.path.join(self.opt_path, 'trials.json'), "w") as fp:
            json.dump(Jsonize(trials.trials)(), fp, sort_keys=True, indent=4, cls=JsonEncoder)

        setattr(self, 'trials', trials)
        # self.results = trials.results
        if self.process_results:
            self._process_results()

        return best

    def _predict(self, *args, **params):

        if self.use_named_args:
            return self.objective_fn(**params)

        if callable(self.objective_fn) and not self.use_named_args:
            return self.objective_fn(*args)

    def hp_space(self) -> dict:
        """returns a dictionary whose values are hyperopt equivalent space instances."""
        return {k: v.as_hp(False if self.algorithm == 'atpe' else True) for k,v in self.space().items()}

    def xy_of_iterations(self) -> dict:
        # todo, not in original order
        if self.backend == "optuna":
            num_iters = range(self.num_iterations)
            return {float(f'{round(trial.value, 5)}_{idx}'): trial.params for idx, trial in zip(num_iters, self.study.trials)}
        elif self.backend == "hyperopt":
            return x_iter_for_tpe(self.trials, self.hp_space(), as_list=False)
        elif self.backend == 'skopt':
            assert self.gpmin_results is not None, f"gpmin_results is not populated yet"
            # adding idx because sometimes the difference between two func_vals is negligible
            fv = self.gpmin_results['func_vals']
            xiters = self.gpmin_results['x_iters']
            return {f'{round(k, 5)}_{idx}': self.to_kw(v) for idx, k,v in zip(range(len(fv)), fv, xiters)}
        else:
            # for sklearn based
            return self.results

    def func_vals(self):
        if self.backend == 'hyperopt':
            return [self.trials.results[i]['loss'] for i in range(self.num_iterations)]
        elif self.backend == 'optuna':
            return [s.values for s in self.study.trials]
        else:
            return np.array(list(self.results.keys()), dtype=np.float32)

    def skopt_results(self):
        if self.use_own and self.algorithm in ["bayes", "bayes_rf"] and self.backend == 'skopt':
            return self.gpmin_results
        else:
            class SR:
                x_iters = [list(s.values()) for s in self.xy_of_iterations().values()]
                func_vals = self.func_vals()
                space = self.skopt_space()
                if isinstance(self.best_paras(), list):
                    x = self.best_paras
                elif isinstance(self.best_paras(), dict):
                    x = list(self.best_paras().values())
                else:
                    raise NotImplementedError

            return SR()

    def best_xy(self) -> dict:
        """Returns best (optimized) parameters as dictionary.
        The key of the dictionary is the value of the objective function with
        best parameters. The value of the dictionary is again a dictionay of
        optimized parameters.
        """
        d = self.xy_of_iterations()
        k = list(dict(sorted(d.items())).keys())[0]
        paras = {k: d[k]}

        return paras

    def _process_results(self):
        """post processing of results"""
        self.save_iterations_as_xy()

        # parallel coordinates of hyperparameters
        d = self.xy_of_iterations()

        data = pd.DataFrame([list(v.values()) for v in d.values()],
                            columns=[s for s in self.space()])
        categories = np.array(list(self.xy_of_iterations().keys())).astype("float64")
        parallel_coordinates(
            data=data,
            categories=categories,
            show=False
        )
        fname = os.path.join(self.opt_path, "parallel_coordinates")
        plt.savefig(fname, dpi=500, bbox_inches="tight")

        # deep learning related results
        if self.objective_fn_is_dl:
            plot_convergences(self.opt_path, what='val_loss', ylabel='Validation MSE')
            plot_convergences(self.opt_path, what='loss', ylabel='MSE', leg_pos="upper right")

        # empirical CDF of objective function
        plt.close("all")
        y = np.array(list(self.xy_of_iterations().keys())).astype("float64")
        plot_edf(y)
        plt.savefig(os.path.join(self.opt_path, "edf"))

        # distributions/historgrams of explored hyperparameters
        self._plot_distributions()

        sr = self.skopt_results()

        # convergence plot
        plt.close('all')
        if sr.x_iters is not None and self.backend != "skopt":
            # todo, should include an option to plot original evaluations instead of only minimum
            plot_convergence([sr])

            fname = os.path.join(self.opt_path, "convergence.png")
            plt.savefig(fname, dpi=300, bbox_inches='tight')

        # plot of hyperparameter space as explored by the optimizer
        if self.backend != 'skopt' and len(self.space()) < 20:  # and len(self.space())>1:
            plt.close('all')
            plot_evaluations(sr, dimensions=self.best_paras(as_list=True))
            plt.savefig(os.path.join(self.opt_path, "evaluations.png"), dpi=300, bbox_inches='tight')

        self.plot_importance(raise_error=False)

        if self.backend == 'hyperopt':
            loss_histogram([y for y in self.trials.losses()],
                           save=True,
                           fname=os.path.join(self.opt_path, "loss_histogram.png")
                           )
            plot_hyperparameters(self.trials,
                                 fname=os.path.join(self.opt_path, "hyperparameters.png"),
                                 save=True)

        if plotly is not None:

            if self.backend == 'optuna':

                fig = plot_contour(self.study)
                plotly.offline.plot(fig, filename=os.path.join(self.opt_path, 'contours.html'), auto_open=False)

        return

    def plot_importance(self, raise_error=True):

        msg = "You must have optuna installed to get hyper-parameter importance."
        if optuna is None:
            if raise_error:
                raise ModuleNotFoundError(msg)
            else:
                warnings.warn(msg)

        else:
            importances, importance_paras, ax = plot_param_importances(self.optuna_study())
            if importances is not None:
                plt.savefig(os.path.join(self.opt_path, 'fanova_importance_bar.png'), bbox_inches="tight", dpi=300)

                plt.close('all')
                df = pd.DataFrame.from_dict(importance_paras)
                axis = df.boxplot(rot=70, return_type="axes")
                axis.set_ylabel("Relative Importance")
                plt.savefig(os.path.join(self.opt_path, "fanova_importance_hist.png"), dpi=300, bbox_inches='tight')

                with open(os.path.join(self.opt_path, "importances.json"), 'w') as fp:
                    json.dump(importances, fp, indent=4, sort_keys=True, cls=JsonEncoder)

                with open(os.path.join(self.opt_path, "fanova_importances.json"), 'w') as fp:
                    json.dump(importance_paras, fp, indent=4, sort_keys=True, cls=JsonEncoder)
        return

    def _plot_distributions(self):
        """plot distributions of explored hyperparameters"""
        from pandas.plotting._matplotlib.tools import create_subplots

        # name of hyperparameters
        h_paras = list(list(self.best_xy().values())[0].keys())

        # container with a list for each hyperparameter
        h_para_lists = {k: [] for k in h_paras}

        for score, x_iter in self.xy_of_iterations().items():
            for para, val in x_iter.items():
                h_para_lists[para].append(val)

        fig, axes = create_subplots(naxes=len(h_paras),
                                    figsize=(6+len(h_paras), 6+len(h_paras)))

        if not isinstance(axes, np.ndarray):
            axes = np.array([axes])

        for ax, col in zip(axes.flat, h_paras):

            hist(h_para_lists[col], ax=ax, show=False,
                 title=col,
                 ylabel="Number of iterations")

        fname = os.path.join(self.opt_path, "distributions.png")
        plt.savefig(fname, bbox_inches="tight")
        return

    def to_kw(self, x):
        names = []
        if isinstance(self.space(), dict):
            for key in self.space().keys():
                names.append(key)
        else:
            raise NotImplementedError

        xkv = {}
        if names is not None:
            for name, val in zip(names, x):
                xkv[name] = val
        else:
            xkv = x

        return xkv

    def eval_with_best(self):
        """
        Find the best parameters and evaluate the objective_fn with them.
        Arguments:
            return_model bool: If True, then then the built objective_fn will be returned
        """

        if self.use_named_args:
            x = self.best_paras()
        else:
            x = self.best_paras(True)

        if self.use_named_args:
            return self.objective_fn(**x)

        if callable(self.objective_fn) and not self.use_named_args:
            if isinstance(x, list) and self.backend == 'hyperopt':  # when x = [x]
                if len(x) == 1:
                    x = x[0]
            return self.objective_fn(x)

        raise NotImplementedError

    @classmethod
    def from_gp_parameters(cls, fpath: str, objective_fn):
        """loads results saved from bayesian optimization"""
        opt_path = os.path.dirname(fpath)
        with open(fpath, 'r') as fp:
            gpmin_results = json.load(fp)
        space = gpmin_results['space']
        spaces = []
        for sp_name, sp_paras in space.items():
            if sp_paras['type'] == 'Categorical':
                spaces.append(Categorical(sp_paras['categories'], name=sp_name))
            elif sp_paras['type'] == 'Integer':
                spaces.append(Integer(low=sp_paras['low'], high=sp_paras['high'], name=sp_name, prior=sp_paras['prior']))
            elif sp_paras['type'] == 'Real':
                spaces.append(Real(low=sp_paras['low'], high=sp_paras['high'], name=sp_name, prior=sp_paras['prior']))
            else:
                raise NotImplementedError

        optimizer = cls('bayes',
                        param_space=spaces,
                        objective_fn=objective_fn,
                        opt_path=opt_path,
                        backend='skopt')
        optimizer.gpmin_results = gpmin_results

        return optimizer

    def pre_calculated_results(self, resutls, from_gp_parameters=True):
        """Loads the pre-calculated results i.e. x and y values which
         have been already evaluated."""
        with open(resutls, 'r') as fp:
            results = json.load(fp)
        return

    def serialize(self):
        return {'fun': '',
                'x': '',
                "best_paras": Jsonize(self.best_paras())(),
                'space': {k: v.serialize() for k, v in self.space().items()},
                'fun_vals': self.func_vals(),
                # 'iters': self.xy_of_iterations(), # todo, for BayesSearchCVs, not getting ys
                'algorithm': self.algorithm,
                'backend': self.backend,
                'opt_path': self.opt_path
                }

    def optuna_study(self) -> Study:
        """
        Attempts to create an optuna Study instance so that
        optuna based plots can be generated.
        Returns None, if not possible."""

        from optuna.trial import TrialState

        if self.backend == 'optuna':
            return self.study

        class _Trial:
            state = TrialState.COMPLETE
            def __init__(self, number:int, values:list, params:dict, distributions:dict):
                values = Jsonize(values)()
                self._number = number
                self._values = values
                if isinstance(values, list):
                    assert len(values) == 1
                    self.value = values[0]
                elif isinstance(values, float) or isinstance(values, int):
                    self.value = values
                else:
                    try:  # try to convert it to float if possible
                        self.value = float(values)
                    except Exception as e:
                        raise NotImplementedError(f"""values must be convertible to list but it is {values} of type
                         {values.__class__.__name__} Actual error message was {e}""")
                self.params = params
                self._distributions = distributions
                self.distributions = distributions

        class _Study(Study):

            trials = []
            idx = 0

            distributions = {sn: s.to_optuna() for sn, s in self.space().items()}

            for _y, _x in self.xy_of_iterations().items():

                assert isinstance(_x, dict), f'params must of type dict but provided params are of type {_x.__class__.__name__}'

                trials.append(_Trial(number=idx,
                                     values=_y,
                                     params=_x,
                                     distributions=distributions
                                     ))
                idx += 1
            best_params = self.best_paras()
            best_trial = None
            best_value = None
            _study_id = 0
            _distributions = distributions

            def __init__(StudyObject):
                pass

            def _is_multi_objective(StudyObject):
                return False

        study = _Study()

        setattr(self, 'study', study)

        return study

    def save_iterations_as_xy(self):

        iterations = self.xy_of_iterations()

        jsonized_iterations = Jsonize(iterations)()

        fname = os.path.join(self.opt_path, "iterations.json")
        with open(fname, "w") as fp:
            json.dump(jsonized_iterations, fp, sort_keys=False, indent=4, cls=JsonEncoder)

        fname = os.path.join(self.opt_path, "iterations_sorted.json")
        with open(fname, "w") as fp:
            json.dump(dict(sorted(jsonized_iterations.items())), fp, sort_keys=True, indent=4, cls=JsonEncoder)

    def add_previous_results(
            self,
            iterations: Union[dict, str] = None,
            x: list = None,
            y :list = None
    ):
        """adds results from previous iterations.

        If you have run the optimization priviously, you can make use
        of those results by appending them.

        Arguments:
            iterations:
                It can be either a dictionary whose keys are y values and values are x
                or it can be a path to a file which contains these xy values as dictioary.
            x:
                a list of lists where each sub-list is the value of hyperparameter
                at at one iteratio. The `x` and `y` arguments optional and will
                only be used if `iterations` are not provided.
            y:
                a list of float values where each value in y is the output
                of objective_fn with corresponding x. The length of `x` and `y`
                must be equal.
        """

        assert self.algorithm in ["bayes", "bayes_rf"]

        if iterations is None:
            assert isinstance(x, list) and isinstance(y, list)
            assert len(x) == len(y), f"x has {len(x)} values while y has {len(y)} values. They must be equal"
            x0 = x
            y0 = y

        elif isinstance(iterations, str):
            assert os.path.exists(iterations), f"the path {iterations} does not exist"
            # it is a path
            with open(iterations, 'r') as fp:
                iter_dict = json.load(fp)

            x0, y0 = self.dict_to_xy(iter_dict)
        else:
            if not isinstance(iterations, dict):
                raise ValueError(f"iterations must be a dictionary but it is of type {iterations.__class__.__name__}")

            x0, y0 = self.dict_to_xy(iterations)

        # todo check for inf and nan in y0

        self.gpmin_args['x0'] = x0
        self.gpmin_args['y0'] = y0

        return

    @staticmethod
    def dict_to_xy(iterations:dict):
        x0, y0 = [], []
        for y, x in iterations.items():
            y0.append(float(y))
            x0.append(list(x.values()))
        return x0, y0



import re

import pandas as pd
import numpy as np


class Resampler(object):
    """Resamples time-series data from one frequency to another frequency.
    """
    min_in_freqs = {
        'MIN': 1,
        'MINUTE': 1,
        'DAILY': 1440,
        'D': 1440,
        'HOURLY': 60,
        'HOUR': 60,
        'H': 60,
        'MONTHLY': 43200,
        'M': 43200,
        'YEARLY': 525600
        }

    def __init__(self, data, freq, how='mean', verbosity=1):
        """
        Arguments:
            data : data to use
            freq : frequency at which to transform/resample
            how : string or dictionary mapping to columns in data defining how to resample the data.
        """
        data = pd.DataFrame(data)
        self.orig_df = data.copy()
        self.target_freq = self.freq_in_mins_from_string(freq)
        self.how = self.check_how(how)
        self.verbosity = verbosity


    def __call__(self, *args, **kwargs):
        if self.target_freq > self.orig_freq:
            # we want to calculate at higher/larger time-step
            return self.downsample()

        else:
            # we want to calculate at smaller time-step
            return self.upsamle()

    @property
    def orig_freq(self):
        return self.freq_in_mins_from_string(pd.infer_freq(self.orig_df.index))

    @property
    def allowed_freqs(self):
        return self.min_in_freqs.keys()

    def check_how(self, how):
        if not isinstance(how, str):
            assert isinstance(how, dict)
            assert len(how) == len(self.orig_df.columns)
        else:
            assert isinstance(how, str)
            how = {col:how for col in self.orig_df.columns}
        return how

    def downsample(self):
        df = pd.DataFrame()
        for col in self.orig_df:
            _df = downsample_df(self.orig_df[col], how=self.how[col], target_freq=self.target_freq)
            df = pd.concat([df, _df], axis=1)

        return df

    def upsamle(self, drop_nan=True):
        df = pd.DataFrame()
        for col in self.orig_df:
            _df = upsample_df(self.orig_df[col], how=self.how[col], target_freq=self.target_freq)
            df = pd.concat([df, _df], axis=1)

        # concatenation of dataframes where one sample was upsampled with linear and the other with same, will result
        # in different length and thus concatenation will add NaNs to the smaller column.
        if drop_nan:
            df = df.dropna()
        return df

    def str_to_mins(self, input_string: str) -> int:

        return self.min_in_freqs[input_string]

    def freq_in_mins_from_string(self, input_string: str) -> int:

        if has_numbers(input_string):
            in_minutes = split_freq(input_string)
        elif input_string.upper() in ['D', 'H', 'M', 'DAILY', 'HOURLY', 'MONTHLY', 'YEARLY', 'MIN', 'MINUTE']:
            in_minutes = self.str_to_mins(input_string.upper())
        else:
            raise TypeError("invalid input string", input_string)

        return int(in_minutes)


def downsample_df(df, how, target_freq):

    if isinstance(df, pd.Series):
        df = pd.DataFrame(df)

    assert how in ['mean', 'sum']
    # from low timestep to high timestep i.e from 1 hour to 24 hour
    # For quantities like temprature, relative humidity, Q, wind speed
    if how == 'mean':
        return df.resample(f'{target_freq}min').mean()
    # For quantities like 'rain', solar radiation', evapotranspiration'
    elif how == 'sum':
        return df.resample(f'{target_freq}min').sum()

def upsample_df(df,  how:str, target_freq:int):
    """drop_nan: if how='linear', we may """
    # from larger timestep to smaller timestep, such as from daily to hourly
    out_freq = str(target_freq) + 'min'

    if isinstance(df, pd.Series):
        df = pd.DataFrame(df)
    col_name  = df.columns[0]
    nan_idx = df.isna()  # preserving indices with nan values
    assert df.shape[1] <=1

    nan_idx_r = nan_idx.resample(out_freq).ffill()
    nan_idx_r = nan_idx_r.fillna(False)  # the first value was being filled with NaN, idk y?
    data_frame = df.copy()

    # For quantities like temprature, relative humidity, Q, wind speed, we would like to do an interpolation
    if how == 'linear':
        data_frame = data_frame.resample(out_freq).interpolate(method='linear')
        # filling those interpolated values with NaNs which were NaN before interpolation
        data_frame[nan_idx_r] = np.nan

    # For quantities like 'rain', solar radiation', evapotranspiration', we would like to distribute them equally
    # at smaller time-steps.
    elif how == 'same':
        # distribute rainfall equally to smaller time steps. like hourly 17.4 will be 1.74 at 6 min resolution
        idx = data_frame.index[-1] + get_offset(data_frame.index.freqstr)
        data_frame = data_frame.append(data_frame.iloc[[-1]].rename({data_frame.index[-1]: idx}))
        data_frame = add_freq(data_frame)
        df1 = data_frame.resample(out_freq).ffill().iloc[:-1]
        df1[col_name ] /= df1.resample(data_frame.index.freqstr)[col_name ].transform('size')
        data_frame = df1.copy()
        # filling those interpolated values with NaNs which were NaN before interpolation
        data_frame[nan_idx_r] = np.nan

    else:
        raise ValueError(f"unoknown method to transform '{how}'")

    return data_frame


def add_freq(df, assert_feq=False, freq=None, method=None):

    idx = df.index.copy()
    if idx.freq is None:
        _freq = pd.infer_freq(idx)
        idx.freq = _freq

        if idx.freq is None:
            if assert_feq:
                df = force_freq(df, freq, method=method)
            else:

                raise AttributeError('no discernible frequency found.  Specify'
                                     ' a frequency string with `freq`.'.format())
        else:
            df.index = idx
    return df


def force_freq(data_frame, freq_to_force, method=None):

    old_nan_counts = data_frame.isna().sum()
    old_shape = data_frame.shape
    dr = pd.date_range(data_frame.index[0], data_frame.index[-1], freq=freq_to_force)

    df_unique = data_frame[~data_frame.index.duplicated(keep='first')]  # first remove duplicate indices if present
    if method:
        df_idx_sorted = df_unique.sort_index()
        df_reindexed = df_idx_sorted.reindex(dr, method='nearest')
    else:
        df_reindexed = df_unique.reindex(dr, fill_value=np.nan)

    df_reindexed.index.freq = pd.infer_freq(df_reindexed.index)
    new_nan_counts = df_reindexed.isna().sum()
    print('Frequency {} is forced to dataframe, NaN counts changed from {} to {}, shape changed from {} to {}'
          .format(df_reindexed.index.freq, old_nan_counts.values, new_nan_counts.values,
                  old_shape, df_reindexed.shape))
    return df_reindexed


def split_freq(freq_str: str) -> int:
    match = re.match(r"([0-9]+)([a-z]+)", freq_str, re.I)
    if match:
        minutes, freq = match.groups()
        if freq.upper() in ['H', 'HOURLY', 'HOURS', 'HOUR']:
            minutes = int(minutes) * 60
        elif freq.upper() in ['D', 'DAILY', 'DAY', 'DAYS']:
            minutes = int(minutes) * 1440
        return int(minutes)
    else:
        raise NotImplementedError

TIME_STEP = {'D': 'Day', 'H': 'Hour', 'M': 'MonthEnd'}

def get_offset(freqstr: str) -> str:
    offset_step = 1
    if freqstr in TIME_STEP:
        freqstr = TIME_STEP[freqstr]
    elif has_numbers(freqstr):
        in_minutes = split_freq(freqstr)
        freqstr = 'Minute'
        offset_step = int(in_minutes)

    offset = getattr(pd.offsets, freqstr)(offset_step)

    return offset

def has_numbers(input_string: str) -> bool:
    return bool(re.search(r'\d', input_string))


import numpy as np

import torch
from torch.utils.data import Dataset


class TorchDataset(Dataset):

    def __init__(self, x, y=None):
        self.x_is_list = False

        if isinstance(x, list):
            self.x_is_list = True

        self.x = x
        self.y = y

        if y is None:
            self.y = torch.tensor(data=np.full(shape=(self.__len__(), 0), fill_value=np.nan))

    def __len__(self):

        if self.x_is_list:
            return len(self.x[0])

        return len(self.x)

    def __getitem__(self, item):

        if self.x_is_list:
            x = [self.x[i][item] for i in range(len(self.x))]
        else:
            x = self.x[item]

        return x, self.y[item]


def to_torch_dataset(x, y=None):
    return TorchDataset(x, y=y)


class TorchMetrics(object):

    def __init__(self, true, predicted):

        self.true = true.view(-1, )
        self.predicted = predicted.view(-1, )

    def r2(self):
        # https://stackoverflow.com/a/66992970/5982232
        target_mean = torch.mean(self.predicted)
        ss_tot = torch.sum((self.predicted - target_mean) ** 2)
        ss_res = torch.sum((self.predicted - self.true) ** 2)
        r2 = 1 - ss_res / ss_tot
        return r2

    def mape(self):
        return (self.predicted - self.true).abs() / (self.true.abs() + 1e-8)

    def nse(self):
        _nse = 1 - sum((self.predicted - self.true) ** 2) / sum((self.true - torch.mean(self.true)) ** 2)
        return _nse

    def pbias(self):
        return 100.0 * sum(self.predicted - self.true) / sum(self.true)

    def mse(self):
        return torch.mean((self.true - self.predicted) ** 2)

    def rmse(self):
        return torch.sqrt(torch.mean(self.true - self.predicted)**2)



__all__ = ["tf", "keras", "torch",
           "xgboost_models", "catboost_models", "lightgbm_models", "sklearn_models",
           "get_attributes"]

from types import FunctionType

try:
    import sklearn
except ModuleNotFoundError:
    sklearn = None


def get_attributes(
        aus,
        what: str,
        retain: str = None,
        case_sensitive: bool = False
) -> dict:
    """gets all callable attributes of aus from what and saves them in dictionary
    with their names as keys. If case_sensitive is True, then the all keys are
    capitalized so that calling them becomes case insensitive. It is possible
    that some of the attributes of tf.keras.layers are callable but still not
    a valid `layer`, sor some attributes of tf.keras.losses are callable but
    still not valid losses, in that case the error will be generated from tensorflow.
    We are not catching those error right now.

    Arguments:
        aus : parent module
        what : child module/package
        retain : if duplicates of 'what' exist then whether to prefer class or function.
            For example, fastica and FastICA exist in sklearn.decomposition then if retain
            is 'function' then fastica will be kept, if retain is 'class' then FastICA is
            kept. If retain is None, then what comes later will overwrite the previously
            kept object.
        case_sensitive : whether to consider what as case-sensitive or not. In such
            a case, fastica and FastICA will both be saved as separate objects.

    Example:
        >>> get_attributes(tf.keras, 'layers')  # will get all layers from tf.keras.layers

    """

    if retain:
        assert retain in ("class", "function")
    all_attrs = {}
    for obj in dir(getattr(aus, what)):
        attr = getattr(getattr(aus, what), obj)
        if callable(attr) and not obj.startswith('_'):

            if not case_sensitive:
                obj = obj.upper()

            if obj in all_attrs and retain == 'function':
                if isinstance(attr, FunctionType):
                    all_attrs[obj] = attr
            elif obj in all_attrs and retain == 'class':
                if not isinstance(attr, FunctionType):
                    all_attrs[obj] = attr
            else:
                all_attrs[obj] = attr

    return all_attrs


def get_sklearn_models():

    if sklearn is not None:
        # the following line must be executed in order for get_attributes to work, don't know why
        from sklearn.ensemble import RandomForestRegressor
        sk_maj_ver = int(sklearn.__version__.split('.')[0])
        sk_min_ver = int(sklearn.__version__.split('.')[1])
        if sk_maj_ver == 0 and sk_min_ver < 24:
            from sklearn.neural_network import multilayer_perceptron
        else:
            from sklearn.neural_network import MLPClassifier
        from sklearn.multioutput import MultiOutputRegressor
        from sklearn.naive_bayes import GaussianNB
        from sklearn.kernel_ridge import KernelRidge
        from sklearn.isotonic import isotonic_regression
        from sklearn.gaussian_process import GaussianProcessRegressor

        from sklearn.experimental import enable_hist_gradient_boosting
        from sklearn.ensemble import HistGradientBoostingClassifier
        from sklearn.ensemble import HistGradientBoostingRegressor
        from sklearn.compose import TransformedTargetRegressor

        skl_models = get_attributes(sklearn, "ensemble", case_sensitive=True)
        skl_models.update(get_attributes(sklearn, "dummy", case_sensitive=True))
        skl_models.update(get_attributes(sklearn, "gaussian_process", case_sensitive=True))
        skl_models.update(get_attributes(sklearn, "compose", case_sensitive=True))
        skl_models.update(get_attributes(sklearn, "linear_model", case_sensitive=True))
        skl_models.update(get_attributes(sklearn, "multioutput", case_sensitive=True))
        skl_models.update(get_attributes(sklearn, "neighbors", case_sensitive=True))
        skl_models.update(get_attributes(sklearn, "neural_network", case_sensitive=True))
        skl_models.update(get_attributes(sklearn, "svm", case_sensitive=True))
        skl_models.update(get_attributes(sklearn, "tree", case_sensitive=True))
        skl_models.update(get_attributes(sklearn, "naive_bayes", case_sensitive=True))
        skl_models.update(get_attributes(sklearn, "kernel_ridge", case_sensitive=True))
        skl_models.update(get_attributes(sklearn, "isotonic", case_sensitive=True))

        from sklearn.calibration import CalibratedClassifierCV
        skl_models.update(get_attributes(sklearn, "calibration", case_sensitive=True))

        from sklearn.semi_supervised import LabelPropagation
        skl_models.update(get_attributes(sklearn, "semi_supervised", case_sensitive=True))

        from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
        skl_models.update(get_attributes(sklearn, "discriminant_analysis", case_sensitive=True))

        skl_models.update({"HistGradientBoostingRegressor": HistGradientBoostingRegressor,
                           "HistGradientBoostingClassifier": HistGradientBoostingClassifier})
    else:
        skl_models = {}

    return skl_models


maj_version = 0
min_version = 0
try:
    from tensorflow import keras
    import tensorflow as tf
    maj_version = int(tf.__version__[0])
    min_version = int(tf.__version__[2])
except ModuleNotFoundError:
    keras = None
    tf = None

try:
    import tcn
except ModuleNotFoundError:
    tcn = None

try:
    import torch
except Exception:  # there can be many reasons for unavailability(unproper installation) fo pytorch
    torch = None

catboost_models = {}

try:
    import catboost
    from catboost import CatBoostClassifier, CatBoostRegressor
    catboost_models.update({"CatBoostClassifier": CatBoostClassifier})
    catboost_models.update({"CatBoostRegressor": CatBoostRegressor})

except ModuleNotFoundError:
    catboost = None


xgboost_models = {}

try:
    import xgboost
    from xgboost import XGBRegressor, XGBClassifier, XGBRFRegressor, XGBRFClassifier
    xgboost_models.update({
        "XGBRegressor": XGBRegressor,
        "XGBClassifier": XGBClassifier,
        "XGBRFRegressor": XGBRFRegressor,
        "XGBRFClassifier": XGBRFClassifier,
    })
except ModuleNotFoundError:
    xgboost = None

lightgbm_models = {}

try:
    import lightgbm
    from lightgbm.sklearn import LGBMClassifier, LGBMRegressor
    lightgbm_models.update({"LGBMClassifier": LGBMClassifier,
                            "LGBMRegressor": LGBMRegressor})
except ModuleNotFoundError:
    lightgbm = None

sklearn_models = get_sklearn_models()

if sklearn is not None:
    from sklearn.experimental import enable_iterative_imputer  # noqa
    imputations = get_attributes(sklearn, 'impute', case_sensitive=True)
else:
    imputations = {}

keras = keras
torch = torch
tf = tf


if tf is not None:
    BACKEND = 'tensorflow'
elif torch is not None:
    BACKEND = 'pytorch'
else:
    BACKEND = None




import os
import numpy as np
import pandas as pd
import matplotlib as mpl
from easy_mpl import regplot
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from sklearn.metrics import plot_roc_curve, plot_confusion_matrix, plot_precision_recall_curve
from SeqMetrics import RegressionMetrics, ClassificationMetrics

from ai4water.utils.visualizations import Plot, init_subplots
from ai4water.utils.utils import dateandtime_now, ts_features, dict_to_file

try:
    import wandb
except ModuleNotFoundError:
    wandb = None

# TODO add Murphy's plot as shown in MLAir
# prediction_distribution aka actual_plot of PDPbox
# average_target_value aka target_plot of PDPbox
# competitive skill score plot/ bootstrap skill score plot as in MLAir
# rank histogram and reliability diagram for probabilitic forecasting model.
# show availability plot of data
# residual plot as in yellow brick
# classification report as in yellow brick
# class prediction error as in yellow brick
# discremination threshold as in yellow brick

class ProcessResults(Plot):
    """post processing of results after training"""
    def __init__(self,
                 forecast_len=None,
                 output_features=None,
                 is_multiclass=None,
                 config: dict = None,
                 path=None,
                 dpi=300,
                 in_cols=None,
                 out_cols=None,
                 verbosity=1,
                 backend: str = 'plotly'
                 ):

        self.forecast_len = forecast_len
        self.output_features = output_features
        self.is_multiclass = is_multiclass
        self.config = config
        self.dpi = dpi
        self.in_cols = in_cols
        self.out_cols = out_cols
        self.quantiles = None
        self.verbosity = verbosity

        super().__init__(path, backend=backend)

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, x):
        self._config = x

    @property
    def quantiles(self):
        return self._quantiles

    @quantiles.setter
    def quantiles(self, x):
        self._quantiles = x

    def horizon_plots(self, errors: dict, fname='', save=True):
        plt.close('')
        _, axis = plt.subplots(len(errors), sharex='all')

        legends = {'r2': "$R^2$", 'rmse': "RMSE", 'nse': "NSE"}
        idx = 0
        for metric_name, val in errors.items():
            ax = axis[idx]
            ax.plot(val, '--o', label=legends.get(metric_name, metric_name))
            ax.legend(fontsize=14)
            if idx >= len(errors)-1:
                ax.set_xlabel("Horizons", fontsize=14)
            ax.set_ylabel(legends.get(metric_name, metric_name), fontsize=14)
            idx += 1
        self.save_or_show(save=save, fname=fname)
        return

    def plot_results(self, true, predicted: pd.DataFrame, save=True, name=None, where=None,
                     annotation_key=None, annotation_val=None, show=False):
        """
        # kwargs can be any/all of followings
            # fillstyle:
            # marker:
            # linestyle:
            # markersize:
            # color:
        """

        regplot(true,
                predicted,
                title=name,
                annotation_key=annotation_key,
                annotation_val=annotation_val,
                show=False
                )

        self.save_or_show(save=save, fname=f"{name}_reg", close=False,
                          where=where, show=show)

        mpl.rcParams.update(mpl.rcParamsDefault)

        _, axis = init_subplots(width=12, height=8)

        # it is quite possible that when data is datetime indexed, then it is not
        # equalidistant and large amount of graph
        # will have not data in that case lines plot will create a lot of useless
        # interpolating lines where no data is present.
        datetime_axis = False
        if isinstance(true.index, pd.DatetimeIndex) and pd.infer_freq(true.index) is not None:
            style = '.'
            true = true
            predicted = predicted
            datetime_axis = True
        else:
            if np.isnan(true.values).sum() > 0:
                style = '.'  # For Nan values we should be using this style otherwise nothing is plotted.
            else:
                style = '-'
                true = true.values
                predicted = predicted.values

        ms = 4 if style == '.' else 2

        if len(true) > 1000:  # because the data is very large, so better to use small marker size
            ms = 2

        axis.plot(predicted, style, color='r',  label='Prediction')

        axis.plot(true, style, color='b', marker='o', fillstyle='none',  markersize=ms, label='True')

        axis.legend(loc="best", fontsize=22, markerscale=4)

        if datetime_axis:
            loc = mdates.AutoDateLocator(minticks=4, maxticks=6)
            axis.xaxis.set_major_locator(loc)
            fmt = mdates.AutoDateFormatter(loc)
            axis.xaxis.set_major_formatter(fmt)

        plt.xticks(fontsize=18)
        plt.yticks(fontsize=18)
        plt.xlabel("Time", fontsize=18)

        self.save_or_show(save=save, fname=name, close=False, where=where)
        return

    def plot_loss(self, history: dict, name="loss_curve", show=False):
        """Considering history is a dictionary of different arrays, possible
        training and validation loss arrays, this method plots those arrays."""

        plt.clf()
        plt.close('all')
        fig = plt.figure()
        plt.style.use('ggplot')
        i = 1

        legends = {
            'mean_absolute_error': 'Mean Absolute Error',
            'mape': 'Mean Absolute Percentage Error',
            'mean_squared_logarithmic_error': 'Mean Squared Logrithmic Error',
            'pbias': "Percent Bias",
            "nse": "Nash-Sutcliff Efficiency",
            "kge": "Kling-Gupta Efficiency",
            "tf_r2": "$R^{2}$",
            "r2": "$R^{2}$"
        }

        sub_plots = {1: {'axis': (1, 1), 'width': 9, 'height': 6},
                     2: {'axis': (1, 2), 'width': 9, 'height': 6},
                     3: {'axis': (1, 3), 'width': 9, 'height': 6},
                     4: {'axis': (2, 2), 'width': 9, 'height': 6},
                     5: {'axis': (5, 1), 'width': 8, 'height': 12},
                     6: {'axis': (3, 2), 'width': 8, 'height': 12},
                     7: {'axis': (3, 2), 'width': 20, 'height': 20},
                     8: {'axis': (4, 2), 'width': 20, 'height': 20},
                     9: {'axis': (5, 2), 'width': 20, 'height': 20},
                     10: {'axis': (5, 2), 'width': 20, 'height': 20},
                     12: {'axis': (4, 3), 'width': 20, 'height': 20},
                     }

        epochs = range(1, len(history['loss']) + 1)
        axis_cache = {}

        for key, val in history.items():

            m_name = key.split('_')[1:] if 'val' in key and '_' in key else key

            if isinstance(m_name, list):
                m_name = '_'.join(m_name)
            if m_name in list(axis_cache.keys()):
                axis = axis_cache[m_name]
                axis.plot(epochs, val, color=[0.96707953, 0.46268314, 0.45772886],
                          label='Validation ')
                axis.legend()
            else:
                axis = fig.add_subplot(*sub_plots[len(history)]['axis'], i)
                axis.plot(epochs, val, color=[0.13778617, 0.06228198, 0.33547859],
                          label='Training ')
                axis.legend()
                axis.set_xlabel("Epochs")
                axis.set_ylabel(legends.get(key, key))
                axis_cache[key] = axis
                i += 1
            axis.set(frame_on=True)

        fig.set_figheight(sub_plots[len(history)]['height'])
        fig.set_figwidth(sub_plots[len(history)]['width'])
        self.save_or_show(fname=name, save=True if name is not None else False, show=show)
        mpl.rcParams.update(mpl.rcParamsDefault)
        return

    def plot_all_qs(self, true_outputs, predicted, save=False):
        plt.close('all')
        plt.style.use('ggplot')

        st, en = 0, true_outputs.shape[0]

        plt.plot(np.arange(st, en), true_outputs[st:en, 0], label="True", color='navy')

        for idx, q in enumerate(self.quantiles):
            q_name = "{:.1f}".format(q * 100)
            plt.plot(np.arange(st, en), predicted[st:en, idx], label="q {} %".format(q_name))

        plt.legend(loc="best")
        self.save_or_show(save, fname="all_quantiles", where='results')

        return

    def plot_quantiles1(self, true_outputs, predicted, st=0, en=None, save=True):
        plt.close('all')
        plt.style.use('ggplot')
        assert true_outputs.shape[-2:] == (1, 1)
        if en is None:
            en = true_outputs.shape[0]
        for q in range(len(self.quantiles) - 1):
            st_q = "{:.1f}".format(self.quantiles[q] * 100)
            en_q = "{:.1f}".format(self.quantiles[-q] * 100)

            plt.plot(np.arange(st, en), true_outputs[st:en, 0], label="True",
                     color='navy')
            plt.fill_between(np.arange(st, en), predicted[st:en, q].reshape(-1,),
                             predicted[st:en, -q].reshape(-1,), alpha=0.2,
                             color='g', edgecolor=None, label=st_q + '_' + en_q)
            plt.legend(loc="best")
            self.save_or_show(save, fname='q' + st_q + '_' + en_q, where='results')
        return

    def plot_quantiles2(self, true_outputs, predicted, st=0, en=None,
                        save=True):
        plt.close('all')
        plt.style.use('ggplot')

        if en is None:
            en = true_outputs.shape[0]
        for q in range(len(self.quantiles) - 1):
            st_q = "{:.1f}".format(self.quantiles[q] * 100)
            en_q = "{:.1f}".format(self.quantiles[q + 1] * 100)

            plt.plot(np.arange(st, en), true_outputs[st:en, 0], label="True",
                     color='navy')
            plt.fill_between(np.arange(st, en),
                             predicted[st:en, q].reshape(-1,),
                             predicted[st:en, q + 1].reshape(-1,),
                             alpha=0.2,
                             color='g', edgecolor=None, label=st_q + '_' + en_q)
            plt.legend(loc="best")
            self.save_or_show(save, fname='q' + st_q + '_' + en_q + ".png",
                              where='results')
        return

    def plot_quantile(self, true_outputs, predicted, min_q: int, max_q, st=0,
                      en=None, save=False):
        plt.close('all')
        plt.style.use('ggplot')

        if en is None:
            en = true_outputs.shape[0]
        q_name = "{:.1f}_{:.1f}_{}_{}".format(self.quantiles[min_q] * 100,
                                              self.quantiles[max_q] * 100, str(st),
                                              str(en))

        plt.plot(np.arange(st, en), true_outputs[st:en, 0], label="True", color='navy')
        plt.fill_between(np.arange(st, en),
                         predicted[st:en, min_q].reshape(-1,),
                         predicted[st:en, max_q].reshape(-1,),
                         alpha=0.2,
                         color='g', edgecolor=None, label=q_name + ' %')
        plt.legend(loc="best")
        self.save_or_show(save, fname="q_" + q_name + ".png", where='results')
        return

    def roc_curve(self, estimator, x, y, save=True):

        if hasattr(estimator, '_model'):
            if estimator._model.__class__.__name__ in ["XGBClassifier", "XGBRFClassifier"] and isinstance(x, np.ndarray):
                x = pd.DataFrame(x, columns=estimator.input_features)

        plot_roc_curve(estimator, x, y.reshape(-1, ))
        self.save_or_show(save, fname="roc", where="results")
        return

    def confusion_matrx(self,estimator, x, y, save=True):

        plot_confusion_matrix(estimator, x, y.reshape(-1, ))
        self.save_or_show(save, fname="confusion_matrix", where="results")
        return

    def precision_recall_curve(self, estimator, x, y, save=True):

        if hasattr(estimator, '_model'):
            if estimator._model.__class__.__name__ in ["XGBClassifier", "XGBRFClassifier"] and isinstance(x, np.ndarray):
                x = pd.DataFrame(x, columns=estimator.input_features)
        plot_precision_recall_curve(estimator, x, y.reshape(-1, ))
        self.save_or_show(save, fname="plot_precision_recall_curve", where="results")
        return

    def process_rgr_results(
            self,
            true: np.ndarray,
            predicted: np.ndarray,
            metrics="minimal",
            prefix=None,
            index=None,
            remove_nans=True,
            user_defined_data: bool = False,
            annotate_with="r2",
    ):
        """
        predicted, true are arrays of shape (examples, outs, forecast_len).
        annotate_with : which value to write on regression plot
        """

        metric_names = {'r2': "$R^2$"}

        if user_defined_data:
            # when data is user_defined, we don't know what out_cols, and forecast_len are
            if predicted.ndim == 1:
                out_cols = ['output']
            else:
                out_cols = [f'output_{i}' for i in range(predicted.shape[-1])]
            forecast_len = 1
            true, predicted = self.maybe_not_3d_data(true, predicted)
        else:
            # for cases if they are 2D/1D, add the third dimension.
            true, predicted = self.maybe_not_3d_data(true, predicted)

            forecast_len = self.forecast_len
            if isinstance(forecast_len, dict):
                forecast_len = np.unique(list(forecast_len.values())).item()

            out_cols = self.output_features
            if isinstance(out_cols, dict):
                _out_cols = []
                for cols in out_cols.values():
                    _out_cols = _out_cols + cols
                out_cols = _out_cols
                if len(out_cols)>1 and not isinstance(predicted, np.ndarray):
                    raise NotImplementedError("""
                    can not process results with more than 1 output arrays""")

        for idx, out in enumerate(out_cols):

            horizon_errors = {metric_name: [] for metric_name in ['nse', 'rmse']}
            for h in range(forecast_len):

                errs = dict()

                fpath = os.path.join(self.path, out)
                if not os.path.exists(fpath):
                    os.makedirs(fpath)

                t = pd.DataFrame(true[:, idx, h], index=index, columns=['true_' + out])
                p = pd.DataFrame(predicted[:, idx, h], index=index, columns=['pred_' + out])

                if wandb is not None and self.config['wandb_config'] is not None:
                    _wandb_scatter(t.values, p.values, out)

                df = pd.concat([t, p], axis=1)
                df = df.sort_index()
                fname = f"{prefix + out}_{h}.csv"
                df.to_csv(os.path.join(fpath, fname), index_label='index')

                annotation_val = getattr(RegressionMetrics(t, p), annotate_with)()
                self.plot_results(t,
                                  p,
                                  name=fname,
                                  where=out,
                                  annotation_key=metric_names.get(annotate_with, annotate_with),
                                  annotation_val=annotation_val,
                                  show=bool(self.verbosity))

                if remove_nans:
                    nan_idx = np.isnan(t)
                    t = t.values[~nan_idx]
                    p = p.values[~nan_idx]

                errors = RegressionMetrics(t, p)
                errs[out + '_errors_' + str(h)] = getattr(errors, f'calculate_{metrics}')()
                errs[out + 'true_stats_' + str(h)] = ts_features(t)
                errs[out + 'predicted_stats_' + str(h)] = ts_features(p)

                dict_to_file(fpath, errors=errs, name=prefix)

                for p in horizon_errors.keys():
                    horizon_errors[p].append(getattr(errors, p)())

            if forecast_len > 1:
                self.horizon_plots(horizon_errors, f'{prefix}_{out}_horizons.png')
        return

    def process_cls_results(
            self,
            true: np.ndarray,
            predicted: np.ndarray,
            metrics="minimal",
            prefix=None,
            index=None,
            user_defined_data: bool = False
    ):
        """post-processes classification results."""
        if user_defined_data:
            return

        if self.is_multiclass:
            if len(predicted) == predicted.size:
                predicted = predicted.reshape(-1, 1)

            pred_labels = [f"pred_{i}" for i in range(predicted.shape[1])]
            true_labels = [f"true_{i}" for i in range(true.shape[1])]
            fname = os.path.join(self.path, f"{prefix}_prediction.csv")
            pd.DataFrame(np.concatenate([true, predicted], axis=1),
                         columns=true_labels + pred_labels, index=index).to_csv(fname)
            class_metrics = ClassificationMetrics(true, predicted, multiclass=True)

            dict_to_file(self.path,
                         errors=class_metrics.calculate_all(),
                         name=f"{prefix}_{dateandtime_now()}.json")
        else:
            if predicted.ndim == 1:
                predicted = predicted.reshape(-1, 1)
            for idx, _class in enumerate(self.output_features):
                _true = true[:, idx]
                _pred = predicted[:, idx]

                fpath = os.path.join(self.path, _class)
                if not os.path.exists(fpath):
                    os.makedirs(fpath)

                class_metrics = ClassificationMetrics(_true, _pred, multiclass=False)
                dict_to_file(fpath,
                             errors=getattr(class_metrics, f"calculate_{metrics}")(),
                             name=f"{prefix}_{_class}_{dateandtime_now()}.json"
                             )

                fname = os.path.join(fpath, f"{prefix}_{_class}.csv")
                array = np.concatenate([_true.reshape(-1, 1), _pred.reshape(-1, 1)], axis=1)
                pd.DataFrame(array, columns=['true', 'predicted'],  index=index).to_csv(fname)

        return

    def maybe_not_3d_data(self, true, predicted):

        forecast_len = self.forecast_len

        if true.ndim < 3:
            if isinstance(forecast_len, dict):
                forecast_len = set(list(forecast_len.values()))
                assert len(forecast_len) == 1
                forecast_len = forecast_len.pop()

            assert forecast_len == 1, f'{forecast_len}'
            axis = 2 if true.ndim == 2 else (1, 2)
            true = np.expand_dims(true, axis=axis)

        if predicted.ndim < 3:
            assert forecast_len == 1
            axis = 2 if predicted.ndim == 2 else (1, 2)
            predicted = np.expand_dims(predicted, axis=axis)

        return true, predicted


def choose_examples(x, examples_to_use, y=None):
    """Chooses exampels from x and y"""
    if isinstance(examples_to_use, int):
        if len(x) == examples_to_use:
            # since using all the examples, don't randomize them
            x, index = x, np.arange(examples_to_use)
        else:
            # randomly chose x values from test_x
            x, index = choose_n_imp_exs(x, examples_to_use, y)
    elif isinstance(examples_to_use, float):
        assert examples_to_use < 1.0
        # randomly choose x fraction from test_x
        x, index = choose_n_imp_exs(x, int(examples_to_use * len(x)), y)

    elif hasattr(examples_to_use, '__len__'):
        index = np.array(examples_to_use)
        x = x[index]
    else:
        raise ValueError(f"unrecognized value of examples_to_use: {examples_to_use}")

    return x, index


def choose_n_imp_exs(x: np.ndarray, n: int, y=None):
    """Chooses the n important examples from x and y"""

    n = min(len(x), n)

    st = n // 2
    en = n - st

    if y is None:
        idx = np.random.randint(0, len(x), n)
    else:
        st = np.argsort(y, axis=0)[0:st].reshape(-1,)
        en = np.argsort(y, axis=0)[-en:].reshape(-1,)
        idx = np.hstack([st, en])

    x = x[idx]

    return x, idx


def _wandb_scatter(true: np.ndarray, predicted: np.ndarray, name: str) -> None:
    """Adds a scatter plot on wandb."""
    data = [[x, y] for (x, y) in zip(true.reshape(-1,), predicted.reshape(-1,))]
    table = wandb.Table(data=data, columns=["true", "predicted"])
    wandb.log({
        "scatter_plot": wandb.plot.scatter(table, "true", "predicted",
                                           title=name)
               })
    return




import os
import warnings

import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
from easy_mpl import bar_chart, imshow

from ai4water.backend import xgboost, tf
from ai4water.utils.visualizations import Plot
from ai4water.utils.utils import plot_activations_along_inputs


class Interpret(Plot):
    """Interprets the ai4water Model."""

    def __init__(self, model):
        """
        Arguments
        ---------
            model :
                an instance of ai4water's Model
        """
        self.model = model

        super().__init__(model.path)

        if self.model.category.upper() == "DL":

            if hasattr(model, 'interpret') and not model.__class__.__name__ == "Model":
                model.interpret()
            else:

                if hasattr(model, 'TemporalFusionTransformer_attentions'):
                    atten_components = self.tft_attention_components()

        elif self.model.category == 'ML':
            use_xgb = False
            if self.model._model.__class__.__name__ == "XGBRegressor":
                use_xgb = True
            self.plot_feature_importance(use_xgb = use_xgb)

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, x):
        self._model = x

    def feature_importance(self):
        if self.model.category.upper() == "ML":

            estimator = self.model._model

            if not is_fitted(estimator):
                print(f"the model {estimator} is not fitted yet so not feature importance")
                return

            model_name = list(self.model.config['model'].keys())[0]
            if model_name.upper() in ["SVC", "SVR"]:
                if estimator.kernel == "linear":
                    # https://stackoverflow.com/questions/41592661/determining-the-most-contributing-features-for-svm-classifier-in-sklearn
                    return estimator.coef_
            elif hasattr(estimator, "feature_importances_"):
                return estimator.feature_importances_

    def f_importances_svm(self, coef, names, save):

        plt.close('all')
        mpl.rcParams.update(mpl.rcParamsDefault)
        classes = coef.shape[0]
        features = coef.shape[1]
        _, axis = plt.subplots(classes, sharex='all')
        axis = axis if hasattr(axis, "__len__") else [axis]

        for idx, ax in enumerate(axis):
            # colors = ['red' if c < 0 else 'blue' for c in self._model.coef_[idx]]
            ax.bar(range(features), self._model.coef_[idx], 0.4)

        plt.xticks(ticks=range(features), labels=self.model.input_features, rotation=90, fontsize=12)
        self.save_or_show(save=save, fname=f"{list(self.model.config['model'].keys())[0]}_feature_importance")
        return

    def plot_feature_importance(self,
                                importance=None,
                                save=True,
                                show=False,
                                use_xgb=False,
                                max_num_features=20,
                                figsize=None,
                                **kwargs):

        figsize = figsize or (8, 8)

        if importance is None:
            importance = self.feature_importance()

        if self.model.category == "ML":
            model_name = list(self.model.config['model'].keys())[0]
            if model_name.upper() in ["SVC", "SVR"]:
                if self.model._model.kernel == "linear":
                    return self.f_importances_svm(importance, self.model.input_features, save=save)
                else:
                    warnings.warn(f"for {self.model._model.kernel} kernels of {model_name}, feature "
                                  f"importance can not be plotted.")
                return

        if isinstance(importance, np.ndarray):
            assert importance.ndim <= 2

        if importance is None:
            return

        all_cols = self.model.input_features + self.model.output_features
        if self.model.teacher_forcing:
            all_cols = self.model.input_features

        imp_sort = np.sort(importance)[::-1]
        all_cols = np.array(all_cols)
        all_cols = all_cols[np.argsort(importance)[::-1]]

        # save the whole importance before truncating it
        fname = os.path.join(self.model.path, 'feature_importance.csv')
        pd.DataFrame(imp_sort, index=all_cols,
                     columns=['importance_sorted']).to_csv(fname)

        imp = np.concatenate([imp_sort[0:max_num_features], [imp_sort[max_num_features:].sum()]])
        all_cols = list(all_cols[0:max_num_features]) + [f'rest_{len(all_cols) - max_num_features}']

        if use_xgb:
            self._feature_importance_xgb(max_num_features=max_num_features,
                                         save=save, show=show)
        else:
            plt.close('all')
            _, axis = plt.subplots(figsize=figsize)
            bar_chart(labels=all_cols,
                      values=imp,
                      ax=axis,
                      title="Feature importance",
                      show=False,
                      xlabel_fs=12)
            self.save_or_show(save=save, show=show, fname="feature_importance.png")
        return

    def _feature_importance_xgb(self, save=True, show=False, max_num_features=None, **kwargs):
        # https://xgboost.readthedocs.io/en/latest/python/python_api.html#xgboost.plot_importance
        if xgboost is None:
            warnings.warn("install xgboost to plot plot_importance using xgboost", UserWarning)
        else:
            booster = self.model._model.get_booster()
            booster.feature_names = self.model.input_features
            plt.close('all')
            # global feature importance with xgboost comes with different types
            xgboost.plot_importance(booster, max_num_features=max_num_features)
            self.save_or_show(save=save, show=show,
                              fname="feature_importance_weight.png")
            plt.close('all')
            xgboost.plot_importance(booster, importance_type="cover",
                                    max_num_features=max_num_features, **kwargs)
            self.save_or_show(save=save, show=show,
                              fname="feature_importance_type_cover.png")
            plt.close('all')
            xgboost.plot_importance(booster, importance_type="gain",
                                    max_num_features=max_num_features, **kwargs)
            self.save_or_show(save=save, show=show,
                              fname="feature_importance_type_gain.png")

        return

    def compare_xgb_f_imp(
            self,
            calculation_method="all",
            rescale=True,
            fig_width=1200,
            fig_height=1000,
    ):
        """compare various feature importance calculations methods that are built
        in in XGBoost"""

        import plotly.graph_objects as go
        from plotly.subplots import make_subplots

        inp_features = self.model.input_features
        assert isinstance(inp_features, list)

        booster = self.model._model.get_booster()
        booster.feature_names = self.model.input_features

        _importance_types = ['weight', 'gain', 'cover', 'total_gain', 'total_cover']
        importance_types = _importance_types.copy()

        if calculation_method != "all":
            if isinstance(calculation_method, str):
                calculation_method = [calculation_method]
            assert isinstance(calculation_method, list)
            # remove those which are not desired
            for imp in _importance_types:
                if imp not in calculation_method:
                    importance_types.remove(imp)

        # container to hold importances with each method
        importance = []

        for idx, imp_type in enumerate(importance_types):
            score = pd.Series(booster.get_score(importance_type=imp_type))
            score = pd.DataFrame(score, columns=[imp_type])

            if rescale:
                # so that the sum of all feature importance is 1.0 and the
                # scale is relative
                score = score / score.sum()

            importance.append(score)

        importance = pd.concat(importance, axis=1)

        # initiate figure with subplots
        fig = make_subplots(
            rows=len(importance_types) + 1, cols=1,
            vertical_spacing=0.02
            # shared_xaxes=True
        )

        for idx, col in enumerate(importance.columns):
            fig.add_trace(go.Bar(
                x=importance.index.tolist(),
                y=importance[col],
                name=col
            ), row=idx + 1, col=1)

        fig.update_xaxes(showticklabels=False)  # hide all the xticks
        fig.update_xaxes(showticklabels=True,
                         row=len(importance_types),
                         col=1,
                         tickangle=-45,
                         title="Input Features"
                         )

        # Here we modify the tickangle of the xaxis, resulting in rotated labels.
        fig.update_layout(
            height=fig_height,
            width=fig_width,
            legend_title="Calculation Method",
            title_text="XGBoost Feature Importance",
            title_x=0.42,
            font=dict(
                family="Times New Roman",
                size=26,
            )
        )
        fname = os.path.join(self.model.path, "xgb_f_imp_comp.html")
        fig.write_html(fname)
        return fig

    def tft_attention_components(
            self,
            data='test'
    ):
        """
        Gets attention components of tft layer from ai4water's Model.

        Arguments:
            data : the data to use to calculate attention components

        Returns
        -------
        dict
            dictionary containing attention components of tft as numpy arrays.
            Following four attention components are present in the dictionary
                - decoder_self_attn: (attention_heads, ?, total_time_steps, 22)
                - static_variable_selection_weights:
                - encoder_variable_selection_weights: (?, encoder_steps, input_features)
                - decoder_variable_selection_weights: (?, decoder_steps, input_features)
        str
            a string indicating which data was used
        """

        maybe_create_path(self.model.path)

        x, _, = getattr(self.model, f'{data}_data')()

        if len(x) == 0 and data == "test":
            warnings.warn("No test data found. using validation data instead",
                          UserWarning)
            data = 'validation'
            x, _, = getattr(self.model, f'{data}_data')()

        attentions = self.model.TemporalFusionTransformer_attentions
        if self.model.api == 'subclassing':
            inputs = self.model.inputs
        else:
            inputs = self.model._model.inputs

        attention_components = {}

        for k, v in attentions.items():
            if v is not None:
                temp_model = tf.keras.Model(inputs=inputs,
                                            outputs=v)
                attention_components[k] = temp_model.predict(x=x, verbose=1, steps=1)

        return attention_components, data

    def get_enc_var_selection_weights(self, data='test'):
        """Returns encoder variable selection weights of TFT model"""

        mpl.rcParams.update(mpl.rcParamsDefault)

        ac, data = self.tft_attention_components(data=data)
        return ac['encoder_variable_selection_weights']


    def interpret_example_tft(self, example_index, model=None, data='test',
                              show=False):
        """interprets a single example using TFT model.

        Parameters
        ---------
            example_index :
                index of example to be explained
            model : t
                he ai4water model
            data :
                the data whose example to interpret.
            show :
                whether to show the plot or not
        """
        if isinstance(data, str):
            assert data in ("training", "test", "validation")
            data_name = data
        else:
            data_name = "data"

        enc_var_selection_weights = self.get_enc_var_selection_weights(data=data)

        plt.close('all')

        axis, im = imshow(enc_var_selection_weights[example_index],
                          aspect="auto",
                          ylabel="lookback steps",
                          title=example_index,
                          show=False
                          )

        plt.xticks(np.arange(model.num_ins), model.input_features, rotation=90)
        plt.colorbar(im, orientation='vertical', pad=0.05)
        plt.savefig(os.path.join(maybe_create_path(model.path), f'{data_name}_enc_var_selec_{example_index}.png'),
                    bbox_inches='tight', dpi=300)
        if show:
            plt.show()
        return

    def interpret_tft(self, data="test"):
        """global interpretation of TFT model.

        Arguments:
            data : the data to use to interpret model
        """

        true, predictions = model.predict(data=data, return_true=True,
                                          process_results=False)

        ac, data = self.tft_attention_components(data=data)

        encoder_variable_selection_weights = ac['encoder_variable_selection_weights']

        train_x, train_y = getattr(model, f'{data}_data')()

        plot_activations_along_inputs(activations=encoder_variable_selection_weights,
                                      data=train_x[:, -1],
                                      observations=true,
                                      predictions=predictions,
                                      in_cols=model.input_features,
                                      out_cols=model.output_features,
                                      lookback=model.lookback,
                                      name=f'tft_encoder_weights_{data}',
                                      path=maybe_create_path(model.path)
                                      )
        return


def maybe_create_path(path):
    path = os.path.join(path, "interpret")
    if not os.path.exists(path):
        os.makedirs(path)
    return path


def is_fitted(estimator):

    if hasattr(estimator, 'is_fitted'):  # for CATBoost
        return estimator.is_fitted

    attrs = [v for v in vars(estimator)
             if v.endswith("_") and not v.startswith("__")]

    if not attrs:
        return False

    return True




from scipy.special import boxcox

from sklearn.preprocessing import MinMaxScaler as SKMinMaxScaler
from sklearn.preprocessing import StandardScaler as SKStandardScaler
from sklearn.preprocessing import RobustScaler as SKRobustScaler
from sklearn.preprocessing import PowerTransformer as SKPowerTransformer
from sklearn.preprocessing import QuantileTransformer as SKQuantileTransformer
from sklearn.preprocessing import FunctionTransformer as SKFunctionTransformer
from sklearn.preprocessing import MaxAbsScaler as SKMaxAbsScaler
from sklearn.utils.validation import check_is_fitted

import numpy as np

from ai4water.utils.utils import jsonize


# todo
# inverse hyperbolic transformation: effective with many zeros

class ScalerWithConfig(object):
    """Extends the sklearn's scalers in such a way that they can be
    saved to a json file an d loaded from a json file

    Methods
    --------
        - config
        - form_config
    """

    @property
    def config_paras(self) -> list:
        raise NotImplementedError

    def get_params(self):
        raise NotImplementedError

    @classmethod
    def from_config(cls, config: dict):
        """Build the scaler/transformer from config

        Arguments:
            config : dictionary of parameters which can be used to build transformer/scaler.

        Returns :
            An instance of scaler/transformer
        """
        scaler = cls(**config['params'])
        setattr(scaler, '_config', config['config'])
        setattr(scaler, '_from_config', True)
        for attr, attr_val in config['config'].items():
            setattr(scaler, attr, attr_val)
        return scaler

    def config(self) -> dict:
        """Returns all the parameters in scaler/transformer in a dictionary"""
        if self.__class__.__name__ == 'MyFunctionTransformer':
            pass
        else:
            check_is_fitted(self)

        _config = {}
        for attr in self.config_paras:
            _config[attr] = getattr(self, attr)

        return {"params": self.get_params(),
                "config": _config}


class MinMaxScaler(SKMinMaxScaler, ScalerWithConfig):

    @property
    def config_paras(self):
        return ['scale_', 'min_', 'n_samples_seen_', 'data_min_', 'data_max_', 'data_range_']


class StandardScaler(SKStandardScaler, ScalerWithConfig):

    @property
    def config_paras(self):
        return ['scale_', 'n_samples_seen_', 'mean_', 'var_']


class RobustScaler(SKRobustScaler, ScalerWithConfig):

    @property
    def config_paras(self):
        return ['scale_', 'center_']


class PowerTransformer(SKPowerTransformer, ScalerWithConfig):
    """This transformation enhances scikit-learn's PowerTransformer by allowing
    the user to define `lambdas` parameter for each input feature. The default
    behaviour of this transformer is same as that of scikit-learn's.
    """
    def __init__(self, method='yeo-johnson', *, standardize=True, copy=True,
                 lambdas=None):
        """
        lambdas: float or 1d array like for each feature. If not given, it is
            calculated from scipy.stats.boxcox(X, lmbda=None). Only available
            if method is box-cox.
        For complete documentation see scikit-learn's documentation_

        .. documentation:
            https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.PowerTransformer.html
        """
        if lambdas is not None:
            if isinstance(lambdas, float):
                lambdas = np.array([lambdas])
            lambdas = np.array(lambdas)
            # if given lambdas must be a 1d array
            assert lambdas.size == len(lambdas)
            lambdas = lambdas.reshape(-1,)
            assert method != "yeo-johnson"

        self.lambdas = lambdas

        super(PowerTransformer, self).__init__(method=method, standardize=standardize,
                                               copy=copy)

    @property
    def config_paras(self):
        return ['lambdas_', 'scaler_to_standardize_']

    @classmethod
    def from_config(cls, config: dict):
        """Build the scaler/transformer from config

        Arguments:
            config : dictionary of parameters which can be used to build transformer/scaler.

        Returns :
            An instance of scaler/transformer
        """
        scaler = cls(**config['params'])
        setattr(scaler, '_config', config['config'])
        setattr(scaler, '_from_config', True)

        _scaler_config = config['config'].pop('scaler_to_standardize_')
        setattr(scaler, '_scaler', StandardScaler.from_config(_scaler_config))

        for attr, attr_val in config['config'].items():
            setattr(scaler, attr, attr_val)
        return scaler

    def _fit(self, X, y=None, force_transform=False):
        """copying from sklearn because we want to use our own StandardScaler
        which can be serialzied. and optionally with user provided with lambda
        parameter."""
        X = self._check_input(X, in_fit=True, check_positive=True,
                              check_method=True)

        if not self.copy and not force_transform:  # if call from fit()
            X = X.copy()  # force copy so that fit does not change X inplace

        optim_function = {'box-cox': self._box_cox_optimize,
                          'yeo-johnson': self._yeo_johnson_optimize
                          }[self.method]
        if self.lambdas is None:
            with np.errstate(invalid='ignore'):  # hide NaN warnings
                self.lambdas_ = np.array([optim_function(col) for col in X.T])
        else:  # take user defined lambdas
            self.lambdas_ = self.lambdas

        if self.standardize or force_transform:
            transform_function = {'box-cox': boxcox,
                                  'yeo-johnson': self._yeo_johnson_transform
                                  }[self.method]
            for i, lmbda in enumerate(self.lambdas_):
                with np.errstate(invalid='ignore'):  # hide NaN warnings
                    X[:, i] = transform_function(X[:, i], lmbda)

        setattr(self, 'scaler_to_standardize_', None)
        if self.standardize:
            self._scaler = StandardScaler(copy=False)
            if force_transform:
                X = self._scaler.fit_transform(X)
            else:
                self._scaler.fit(X)

            setattr(self, 'scaler_to_standardize_', self._scaler.config())

        return X

class QuantileTransformer(SKQuantileTransformer, ScalerWithConfig):

    @property
    def config_paras(self):
        return ['n_quantiles_', 'references_', 'quantiles_']


class MaxAbsScaler(SKMaxAbsScaler, ScalerWithConfig):

    @property
    def config_paras(self):
        return ['scale_', 'n_samples_seen_', 'max_abs_']


class Center(ScalerWithConfig):

    def __init__(
            self,
            feature_dim="2d"
    ):
        self.feature_dim = feature_dim

    def fit(self): pass

    def fit_transform(self, x:np.ndarray)->np.ndarray:
        dim = x.ndim
        assert dim == 2

        mean = np.mean(x, axis=0)

        setattr(self, 'mean_', mean)
        setattr(self, 'data_dim_', dim)

        return x - mean

    def inverse_transform(self, x:np.ndarray)->np.ndarray:

        assert x.ndim == self.data_dim_
        return x + self.mean_

    @property
    def config_paras(self):
        return ['data_dim_', 'mean_']

    def get_params(self):
        return {'feature_dim': self.feature_dim}


class FuncTransformer(ScalerWithConfig):

    def __init__(
            self,
            feature_dim: str = "2d"
    ):
        """
        Arguments:
            feature_dim:
                whether the features are 2 dimensional or 1 dimensional. Only
                relevant if the `x` to `fit_transform` is 3D. In such as case
                if feature_dim is `1D`, it will be considered that the x consists
                of following shape (num_examples, time_steps, num_features)

        """
        assert feature_dim in ("1d", "2d")
        self.feature_dim = feature_dim

    @property
    def func(self):
        raise NotImplementedError

    @property
    def inv_func(self):
        raise NotImplementedError

    def fit(self):
        return

    def _get_dim(self, x:np.ndarray):
        dim = x.ndim
        setattr(self, 'data_dim_', dim)

        if dim > 3:
            raise ValueError(f" dimension {dim} not allowed")
        return dim

    def fit_transform(self, x:np.ndarray)-> np.ndarray:

        dim = self._get_dim(x)

        if dim == 3 and self.feature_dim == "1d":
            _x = np.full(x.shape, np.nan)
            for time_step in range(x.shape[1]):
                _x[:, time_step] = self.func(x[:, time_step])
        else:
            _x = self.func(x)

        return _x

    def inverse_transform_without_fit(self, x):
        return self._inverse_transform(x, False)

    def _inverse_transform(self, x, check_dim=True):
        dim = x.ndim
        if check_dim:
            assert dim == self.data_dim_, f"dimension of data changed from {self.data_dim_} to {dim}"

        if dim == 3 and self.feature_dim == "1d":
            _x = np.full(x.shape, np.nan)
            for time_step in range(x.shape[1]):
                _x[:, time_step] = self.inv_func(x[:, time_step])

        elif 2 <= dim < 4:
            _x = self.inv_func(x)
        else:
            raise  ValueError(f" dimension {dim} not allowed")

        return _x

    def inverse_transform(self, x):
        return self._inverse_transform(x)

    @property
    def config_paras(self):
        return ['data_dim_']

    def get_params(self):
        return {'feature_dim': self.feature_dim}


class SqrtScaler(FuncTransformer):

    @property
    def func(self):
        return np.sqrt

    @property
    def inv_func(self):
        return np.square


class LogScaler(FuncTransformer):

    @property
    def func(self):
        return np.log

    @property
    def inv_func(self):
        return np.exp


class Log2Scaler(FuncTransformer):

    @property
    def func(self):
        return np.log2

    @property
    def inv_func(self):
        return lambda x: np.power(2, x)


class Log10Scaler(FuncTransformer):

    @property
    def func(self):
        return np.log10

    @property
    def inv_func(self):
        return lambda x: np.power(10, x)


class TanScaler(FuncTransformer):

    @property
    def func(self):
        return np.tan

    @property
    def inv_func(self):
        return np.tanh


class CumsumScaler(FuncTransformer):

    def fit_transform(self, x:np.ndarray) -> np.ndarray:

        dim = self._get_dim(x)

        if dim == 3 and self.feature_dim == "1d":
            _x = np.full(x.shape, np.nan)
            for time_step in range(x.shape[1]):
                _x[:, time_step] = self.func(x[:, time_step], axis=0)
        else:
            _x = np.cumsum(x, axis=0)

        return _x

    def inverse_transform(self, x):

        dim = x.ndim
        assert dim == self.data_dim_, f"dimension of data changed from {self.data_dim_} to {dim}"

        if dim == 3 and self.feature_dim == "1d":
            _x = np.full(x.shape, np.nan)
            for time_step in range(x.shape[1]):
                _x[:, time_step] = np.diff(x[:, time_step], axis=0, append=0)

        elif 2 <= dim < 4:
            _x = np.diff(x, axis=0, append=0)
        else:
            raise  ValueError(f" dimension {dim} not allowed")

        return _x

class FunctionTransformer(SKFunctionTransformer):
    """Serializing a custom func/inverse_func is difficult. Therefore
    we expect the func/inverse_func to be either numpy function or
    the code as a string.

    Methods
    -------
    from_config

    Attributes
    ----------
    inverse_func_ser

    Example
    -------
        >>> array = np.random.randint(1, 100, (20, 2))
        >>> transformer = FunctionTransformer(func=np.log2,
        >>>                inverse_func="lambda _x: 2**_x", validate=True)
        >>> t_array = transformer.fit_transform(array)
        >>> transformer.config()
        >>> new_transformer = FunctionTransformer.from_config(transformer.config())
        >>> original_array = new_transformer.inverse_transform(t_array)

    """
    def __init__(self, func=None, inverse_func=None, validate=False,
                 accept_sparse=False, check_inverse=True, kw_args=None,
                 inv_kw_args=None):

        # if inverse_func is string, we save a serialized version of it in memory
        # to save it in config later.
        self.inverse_func_ser = inverse_func

        super().__init__(func=func,
                         inverse_func=inverse_func,
                         validate=validate,
                         accept_sparse=accept_sparse,
                         check_inverse=check_inverse,
                         kw_args=kw_args,
                         inv_kw_args=inv_kw_args)

    @property
    def inverse_func(self):
        return self._inverse_func

    @inverse_func.setter
    def inverse_func(self, func):
        self._inverse_func = self.deserialize_func(func)

    @property
    def inverse_func_ser(self):
        return self._inverse_func_ser

    @inverse_func_ser.setter
    def inverse_func_ser(self, func):
        self._inverse_func_ser = self.serialize_func(func)

    @classmethod
    def from_config(cls, config: dict):
        """Build the estimator from config file"""

        func = cls.deserialize_func(config.pop('func'))

        # do not deserialize inverse_func here, it will be done in init method
        scaler = cls(func=func, inverse_func=config.pop('inverse_func'), **cls.deserialize(**config))

        setattr(scaler, '_from_config', True)

        return scaler

    @staticmethod
    def deserialize_func(func):
        if func is not None:
            if isinstance(func, str):
                if func in np.__dict__:
                    func = getattr(np, func)
                else:
                    func = eval(func)
            elif isinstance(func, np.ufunc):  # np.log2
                func = func
            elif func.__name__ in np.__dict__:  # np.diff
                func = func
            else:
                raise ValueError(f"{func}")

        return func

    def config(self) -> dict:
        """Returns all the parameters in scaler in a dictionary"""

        params = self.get_params()
        _config = dict()
        _config['func'] = self.serialize_func(self.func)
        _config['inverse_func'] = self.inverse_func_ser
        _config['kw_args'] = jsonize(self.kw_args)
        _config['inv_kw_args'] = jsonize(self.inv_kw_args)

        for k, v in params.items():
            if k not in _config:
                _config.update({k: v})

        return _config

    @staticmethod
    def deserialize(**kwargs):
        _kwargs = {}
        for k, v in kwargs.items():
            if v == "None":
                v = None
            _kwargs[k] = v

        return _kwargs

    @staticmethod
    def serialize_func(func):

        if type(func) == np.ufunc:
            func = func.__name__

        elif func.__class__.__name__ == "function" and func.__module__ == "numpy":
            func = func.__name__

        elif func is not None:
            if isinstance(func, str):
                func = f"""{func}"""
            else:
                raise ValueError(f"{func} is not serializable")

        return func



# -*- coding: utf-8 -*-
"""
The code in this file has been modified after https://github.com/pangaea-data-publisher/pangaeapy/blob/master/pangaeapy/pandataset.py
"""
import json
import os
import xml.etree.ElementTree as ET
import re
import io

try:
    import requests
except ModuleNotFoundError:
    requests = None

import pandas as pd


class PanParam:
    """ PANGAEA Parameter
    Shoud be used to create PANGAEA parameter objects. Parameter is used here to represent 'measured variables'

    Attributes
    ----------
    ID : int
        the identifier for the parameter
    name : str
        A long name or title used for the parameter
    shortName : str
        A short name or label to identify the parameter
    synonym : dict
        A diconary of synonyms for the parameter whcih e.g. is used by other archives or communities.
        The dict key indicates the namespace (possible values currently are CF and OS)
    type : str
        indicates the data type of the parameter (string, numeric, datetime etc..)
    source : str
        defines the category or source for a parameter (e.g. geocode, data, event)... very PANGAEA specific ;)
    unit : str
        the unit of measurement used with this parameter (e.g. m/s, kg etc..)
    format: str
        the number format string given by PANGAEA e.g ##.000 which defines the displayed precision of the number
    """

    def __init__(self, ID, name, shortName, param_type, source, unit=None, fmt=None):
        self.ID = ID
        self.name = name
        self.shortName = shortName
        # Synonym namespace dict predefined keys are CF: CF variables (), OS:OceanSites, SD:SeaDataNet abbreviations (TEMP, PSAL etc..)
        ns = ('CF', 'OS', 'SD')
        self.synonym = dict.fromkeys(ns)
        self.type = param_type
        self.source = source
        self.unit = unit
        self.format = fmt


class PanDataSet:
    """ PANGAEA DataSet
    The PANGAEA PanDataSet class enables the creation of objects which hold the necessary information, including data as well as metadata, to analyse a given PANGAEA dataset.

    Parameters
    ----------
    ID : str
        The identifier of a PANGAEA dataset. An integer number or a DOI is accepted here
    deleteFlag : str
        in case quality flags are avialable, this parameter defines a flag for which data should not be included in the data dataFrame.
        Possible values are listed here: https://wiki.pangaea.de/wiki/Quality_flag
    addQC : boolean
        adds a QC column for each parameter which contains QC flags

    Attributes
    ----------
    ID : str
        The identifier of a PANGAEA dataset. An integer number or a DOI is accepted here
    params : list of PanParam
        a list of all PanParam objects (the parameters) used in this dataset
    events : list of PanEvent
        a list of all PanEvent objects (the events) used in this dataset
    projects : list of PanProject
        a list containing the PanProjects objects referenced by this dataset
    data : pandas.DataFrame
        a pandas dataframe holding all the data
    loginstatus : str
        a label which indicates if the data set is protected or not default value: 'unrestricted'
    """

    def __init__(self, ID=None, paramlist=None, deleteFlag='', addQC=False):
        ### The constructor allows the initialisation of a PANGAEA dataset object either by using an integer dataset ID or a DOI
        self.ID = setID(ID)
        self.ns = {'md': 'http://www.pangaea.de/MetaData'}
        # Mapping should be moved to e.g netCDF class/module??
        # moddir = os.path.dirname(os.path.abspath(__file__))
        # self.CFmapping=pd.read_csv(moddir+'\\PANGAEA_CF_mapping.txt',delimiter='\t',index_col='ID')
        self.params = dict()
        self.defaultparams = ['Latitude', 'Longitude', 'Event', 'Elevation', 'Date/Time']
        self.paramlist = paramlist
        self.paramlist_index = []
        self.events = []
        self.projects = []
        # allowed geocodes for netcdf generation which are used as xarray dimensions not needed in the moment
        self._geocodes = {1599: 'Date_Time', 1600: 'Latitude', 1601: 'Longitude', 1619: 'Depth water'}
        self.loginstatus = 'unrestricted'
        self.deleteFlag = deleteFlag
        self.metadata = {}
        # print('trying to load data and metadata from PANGAEA')
        self.set_metadata()
        self.defaultparams = [s for s in self.defaultparams if s in self.params.keys()]
        self.addQC = addQC
        if self.paramlist is not None:
            if len(self.paramlist) != len(self.paramlist_index):
                raise ValueError


    def _setParameters(self, panXMLMatrixColumn):
        """
        Initializes the list of parameter objects from the metadata XML info
        """
        coln = dict()
        if panXMLMatrixColumn is not None:
            for matrix in panXMLMatrixColumn:
                paramstr = matrix.find("md:parameter", self.ns)
                panparID = int(_getID(str(paramstr.get('id'))))

                panparShortName = ''
                if paramstr.find('md:shortName', self.ns) is not None:
                    panparShortName = paramstr.find('md:shortName', self.ns).text
                    # Rename duplicate column headers
                    if panparShortName in coln:
                        coln[panparShortName] += 1
                        panparShortName = panparShortName + '_' + str(coln[panparShortName])
                    else:
                        coln[panparShortName] = 1
                panparType = matrix.get('type')
                panparUnit = None
                if paramstr.find('md:unit', self.ns) is not None:
                    panparUnit = paramstr.find('md:unit', self.ns).text
                panparFormat = matrix.get('format')

                self.params[panparShortName] = PanParam(panparID, paramstr.find('md:name', self.ns).text,
                                                        panparShortName, panparType, matrix.get('source'), panparUnit,
                                                        panparFormat)

    def get_data(self, addQC=False):
        """
        This method populates the data DataFrame with data from a PANGAEA dataset.
        In addition to the data given in the tabular ASCII file delivered by PANGAEA.

        Parameters:
        -----------
        addQC : boolean
            If this is set to True, pangaeapy adds a QC column in which the quality flags are separated.
            Each new column is named after the orgininal column plus a "_qc" suffix.
        """

        if self.metadata['hierarchyLevel'] is not None:
            if self.metadata['hierarchyLevel'].get('value') == 'parent':
                raise ValueError(f"""
        Data set is of type parent, please select one of its child datasets.
        The {len(self.children())} child datasets are \n{self.children()}""")

        # converting list of parameters` short names (from user input) to the list of parameters` indexes
        # the list of parameters` indexes is an argument for pd.read_csv
        if self.paramlist is not None:
            self.paramlist += self.defaultparams
            for parameter in self.paramlist:
                _iter = 0
                for shortName in self.params.keys():
                    if parameter == shortName:
                        self.paramlist_index.append(_iter)
                    _iter += 1
            if len(self.paramlist) != len(self.paramlist_index):
                raise ValueError("Error entering parameters`short names!")
        else:
            self.paramlist_index = None

        dataURL = "https://doi.pangaea.de/10.1594/PANGAEA." + str(self.ID) + "?format=textfile"
        panDataTxt = requests.get(dataURL).text
        panData = re.sub(r"/\*(.*)\*/", "", panDataTxt, 1, re.DOTALL).strip()
        # Read in PANGAEA Data
        data = pd.read_csv(io.StringIO(panData), index_col=False, error_bad_lines=False, sep=u'\t',
                                usecols=self.paramlist_index, names=list(self.params.keys()), skiprows=[0])
        # add geocode/dimension columns from Event

        # -- delete values with given QC flags
        if self.deleteFlag != '':
            if self.deleteFlag == '?' or self.deleteFlag == '*':
                self.deleteFlag = "\\" + self.deleteFlag
            data.replace(regex=r'^' + self.deleteFlag + '{1}.*', value='', inplace=True)
            # --- Replace Quality Flags for numeric columns
        if not addQC:
            data.replace(regex=r'^[\?/\*#\<\>]', value='', inplace=True)
        # --- Delete empty columns
        data = data.dropna(axis=1, how='all')
        for paramcolumn in list(self.params.keys()):
            if paramcolumn not in data.columns:
                del self.params[paramcolumn]
            # --- add QC columns
            elif addQC:
                if self.params[paramcolumn].type == 'numeric':
                    data[[paramcolumn + '_qc', paramcolumn]] = data[paramcolumn].astype(str).str.extract(
                        r'(^[\*/\?])?(.+)')
        # --- Adjust Column Data Types
        data = data.apply(pd.to_numeric, errors='ignore')
        if 'Date/Time' in data.columns:
            data.index = pd.to_datetime(data['Date/Time'], format='%Y/%m/%dT%H:%M:%S')
            data.index.names = ['index']
            data.pop('Date/Time')
        return data

    def download(self, path, name=None, **kwargs):

        if name is None:
            name = self.metadata['title'].replace(' ', '_')
            name = name.replace('-', '_')
        path = os.path.join(path, name+'.txt')
        self.get_data().to_csv(path, **kwargs)

        if 'hierarchyLevel' in self.metadata and self.metadata['hierarchyLevel'] is not None:
            self.metadata['hierarchyLevel'] = self.metadata['hierarchyLevel'].text
        fname = os.path.join(os.path.dirname(path), f'{name}_metadata.json')
        with open(fname, 'w') as fp:
            json.dump(self.metadata, fp, indent=4, sort_keys=False)

        return name


    def set_metadata(self):
        """
        The method initializes the metadata of the PanDataSet object using the information of a PANGAEA metadata XML file.
        """
        _metadata = {}
        metaDataURL = "https://doi.pangaea.de/10.1594/PANGAEA." + str(self.ID) + "?format=metainfo_xml"
        r = requests.get(metaDataURL)
        if r.status_code != 404:
            try:
                r.raise_for_status()
                xmlText = r.text
                xml = ET.fromstring(xmlText)
                self.metadata['loginstatus'] = xml.find('./md:technicalInfo/md:entry[@key="loginOption"]', self.ns).get('value')
                if self.metadata['loginstatus'] != 'unrestricted':
                    raise ValueError('Data set is protected')

                self.metadata['hierarchyLevel'] = xml.find('./md:technicalInfo/md:entry[@key="hierarchyLevel"]', self.ns)

                self.metadata['title'] = xml.find("./md:citation/md:title", self.ns).text
                self.metadata['year'] = xml.find("./md:citation/md:year", self.ns).text
                self.metadata['author_info'] = self.find_author_info(xml)
                self.metadata['project_info'] = self.find_project_info(xml)
                self.metadata['license_info'] = self.find_license_info(xml)

                topotypeEl = xml.find("./md:extent/md:topoType", self.ns)
                if topotypeEl is not None:
                    self.topotype = topotypeEl.text
                else:
                    self.topotype = None

                panXMLMatrixColumn = xml.findall("./md:matrixColumn", self.ns)
                self._setParameters(panXMLMatrixColumn)

            except requests.exceptions.HTTPError as e:
                print(e)

    def children(self):
        """Finds the child datasets of a parent dataset"""
        kinder = []
        childqueryURL = "https://www.pangaea.de/advanced/search.php?q=incollection:" + str(self.ID) + "&count=1000"
        r = requests.get(childqueryURL)
        if r.status_code != 404:
            s = r.json()
            for p in s['results']:
                kinder.append(p['URI'])
        return kinder

    def find_license_info(self, xml)->dict:
        lizenz = {}
        idx = 0
        for _license in xml.findall("./md:license", self.ns):

            l = _license.find("md:label", self.ns)
            lizenz[f'label_{idx}'] = l.text if l is not None else l
            n = _license.find("md:name", self.ns)
            lizenz[f'name_{idx}'] = n.text if n is not None else n
            u = _license.find("md:URI", self.ns)
            lizenz[f'URI_{idx}'] = u.text if u is not None else u

            idx += 1

        return lizenz

    def find_project_info(self, xml)->dict:
        projekt_info = {}
        idx = 0
        for project in xml.findall("./md:project", self.ns):

            l = project.find("md:label", self.ns)
            projekt_info[f'label_{idx}'] = l.text if l is not None else l
            n = project.find("md:name", self.ns)
            projekt_info['name_{idx}'] = n.text if n is not None else n
            u = project.find("md:URI", self.ns)
            projekt_info[f'URI_{idx}'] = u.text if u is not None else u
            uri = project.find("md:award/md:URI", self.ns)
            projekt_info[f'awardURI_{idx}'] = uri.text if uri is not None else uri

            idx += 1

        return projekt_info

    def find_author_info(self, xml)->dict:
        autor = {}
        idx = 0

        for author in xml.findall("./md:citation/md:author", self.ns):

            autor[f'lastname_{idx}'] = author.find("md:lastName", self.ns).text
            autor[f'firstname_{idx}'] = author.find("md:firstName", self.ns).text
            orcid = author.find("md:orcid", self.ns)
            autor[f'orcid_{idx}'] = orcid.text if orcid is not None else orcid

            idx += 1

        return autor


def setID(ID):
    """
    Initialize the ID of a data set in case it was not defined in the constructur
    Parameters
    ----------
    ID : str
        The identifier of a PANGAEA dataset. An integer number or a DOI is accepted here
    """
    idmatch = re.search(r'10\.1594\/PANGAEA\.([0-9]+)$', ID)

    if idmatch is not None:
        return idmatch[1]
    else:
       return ID

def _getID(panparidstr):
    panparidstr = panparidstr[panparidstr.rfind('.') + 1:]
    panparId = re.match(r"([a-z]+)([0-9]+)", panparidstr)
    if panparId:
        return panparId.group(2)
    else:
        return False

if __name__ == "__main__":
    # ds = PanDataSet('10.1594/PANGAEA.898217')
    # print(ds.data.shape)
    # ds = PanDataSet('10.1594/PANGAEA.882613')
    # print(ds.data.shape)
    # ds = PanDataSet('10.1594/PANGAEA.879494')
    # print(ds.data.shape)
    # ds = PanDataSet('10.1594/PANGAEA.831196')
    # print(ds.data.shape)
    ds = PanDataSet('10.1594/PANGAEA.919103')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.919104')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.909880')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.908290')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.892384')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.883587')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.882178')
    print(ds.get_data().shape)
    # ds = PanDataSet('10.1594/PANGAEA.811992')
    # print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.807883')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.778629')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.774595')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.746240')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.226925')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.905446')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.900958')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.890070')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.882611')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.879507')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.842446')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.841977')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.831193')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.811072')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.811076')
    print(ds.get_data().shape)
    ds = PanDataSet('10.1594/PANGAEA.912582')
    print(ds.get_data().shape)



import os
import ssl

import tempfile
import sys, shutil
import urllib.request as ulib
import urllib.parse as urlparse

try:
    import requests
except ModuleNotFoundError:
    requests = None


# following files must exist withing data folder for CAMELS-GB data
DATA_FILES = {
    'CAMELS-GB': [
        'CAMELS_GB_climatic_attributes.csv',
        'CAMELS_GB_humaninfluence_attributes.csv',
        'CAMELS_GB_hydrogeology_attributes.csv',
        'CAMELS_GB_hydrologic_attributes.csv',
        'CAMELS_GB_hydrometry_attributes.csv',
        'CAMELS_GB_landcover_attributes.csv',
        'CAMELS_GB_soil_attributes.csv',
        'CAMELS_GB_topographic_attributes.csv'
    ],
    'HYSETS': [  # following files must exist in a folder containing HYSETS dataset.
        'HYSETS_2020_ERA5.nc',
        'HYSETS_2020_ERA5Land.nc',
        'HYSETS_2020_ERA5Land_SWE.nc',
        'HYSETS_2020_Livneh.nc',
        'HYSETS_2020_nonQC_stations.nc',
        'HYSETS_2020_SCDNA.nc',
        'HYSETS_2020_SNODAS_SWE.nc',
        'HYSETS_elevation_bands_100m.csv',
        'HYSETS_watershed_boundaries.zip',
        'HYSETS_watershed_properties.txt'
    ]
}


def download_all_http_directory(url, outpath=None, filetypes=".zip", match_name=None):
    """
    Download all the files which are of category filetypes at the location of
    outpath. If a file is already present. It will not be downloaded.
    filetypes str: extension of files to be downloaded. By default only .zip files
        are downloaded.
    mathc_name str: if not None, then only those files will be downloaded whose name
        have match_name string in them.
    """
    import bs4
    if os.name == 'nt':
        ssl._create_default_https_context = ssl._create_unverified_context
    page = list(urlparse.urlsplit(url))[2].split('/')[-1]
    basic_url = url.split(page)[0]

    r = requests.get(url)
    data = bs4.BeautifulSoup(r.text, "html.parser")
    match_name = filetypes if match_name is None else match_name

    for l in data.find_all("a"):

        if l["href"].endswith(filetypes) and match_name in l['href']:
            _outpath = outpath
            if outpath is not None:
                _outpath = os.path.join(outpath, l['href'])

            if os.path.exists(_outpath):
                print(f"file {l['href']} already exists at {outpath}")
                continue
            download(basic_url + l["href"], _outpath)
            print(r.status_code, l["href"], )


def download(url, out=None):
    """High level function, which downloads URL into tmp file in current
    directory and then renames it to filename autodetected from either URL
    or HTTP headers.

    :param url:
    :param out: output filename or directory
    :return:    filename where URL is downloaded to
    """
    # detect of out is a directory
    if out is not None:
        outdir = os.path.dirname(out)
        out_filename = os.path.basename(out)
        if not os.path.exists(outdir):
            os.makedirs(outdir)
    else:
        outdir = os.getcwd()
        out_filename = None

    # get filename for temp file in current directory
    prefix = filename_from_url(url)
    (fd, tmpfile) = tempfile.mkstemp(".tmp", prefix=prefix, dir=".")
    os.close(fd)
    os.unlink(tmpfile)

    # set progress monitoring callback
    def callback_charged(blocks, block_size, total_size):
        # 'closure' to set bar drawing function in callback
        callback_progress(blocks, block_size, total_size, bar_function=bar)

    callback = callback_charged

    # Python 3 can not quote URL as needed
    binurl = list(urlparse.urlsplit(url))
    binurl[2] = urlparse.quote(binurl[2])
    binurl = urlparse.urlunsplit(binurl)

    (tmpfile, headers) = ulib.urlretrieve(binurl, tmpfile, callback)
    filename = filename_from_url(url)

    if out_filename:
        filename = out_filename

    filename = outdir + "/" + filename

    # add numeric ' (x)' suffix if filename already exists
    if os.path.exists(filename):
        filename = filename + '1'
    shutil.move(tmpfile, filename)

    # print headers
    return filename


__current_size = 0


def callback_progress(blocks, block_size, total_size, bar_function):
    """callback function for urlretrieve that is called when connection is
    created and when once for each block

    draws adaptive progress bar in terminal/console

    use sys.stdout.write() instead of "print,", because it allows one more
    symbol at the line end without linefeed on Windows

    :param blocks: number of blocks transferred so far
    :param block_size: in bytes
    :param total_size: in bytes, can be -1 if server doesn't return it
    :param bar_function: another callback function to visualize progress
    """
    global __current_size

    width = 100

    if sys.version_info[:3] == (3, 3, 0):  # regression workaround
        if blocks == 0:  # first call
            __current_size = 0
        else:
            __current_size += block_size
        current_size = __current_size
    else:
        current_size = min(blocks * block_size, total_size)
    progress = bar_function(current_size, total_size, width)
    if progress:
        sys.stdout.write("\r" + progress)


def filename_from_url(url):
    """:return: detected filename as unicode or None"""
    # [ ] test urlparse behavior with unicode url
    fname = os.path.basename(urlparse.urlparse(url).path)
    if len(fname.strip(" \n\t.")) == 0:
        return None
    return fname


def bar(current_size, total_size, width):
    percent = current_size/total_size * 100
    if round(percent % 1, 4) == 0.0:
        print(f"{round(percent)}% of {round(total_size*1e-6, 2)} MB downloaded")
    return


def check_attributes(attributes, check_against: list) -> list:
    if attributes == 'all' or attributes is None:
        attributes = check_against
    elif not isinstance(attributes, list):
        assert isinstance(attributes, str)
        assert attributes in check_against
        attributes = [attributes]
    else:
        assert isinstance(attributes, list), f'unknown attributes {attributes}'

    assert all(elem in check_against for elem in attributes)

    return attributes


def sanity_check(dataset_name, path):
    if dataset_name in DATA_FILES:
        if dataset_name == 'CAMELS-GB':
            if not os.path.exists(os.path.join(path, 'data')):
                raise FileNotFoundError(f"No folder named `data` exists inside {path}")
            else:
                data_path = os.path.join(path, 'data')
                for file in DATA_FILES[dataset_name]:
                    if not os.path.exists(os.path.join(data_path, file)):
                        raise FileNotFoundError(f"File {file} must exist inside {data_path}")
    return




from typing import Union, List, Dict

import numpy as np
import pandas as pd

from ai4water.utils.utils import jsonize, deepcopy_dict_without_clone
from ai4water.preprocessing.transformations import Transformation
from .utils import TransformerNotFittedError, SP_METHODS


class Transformations(object):
    """
    While the [Transformation][ai4water.preprocessing.transformations.Transformation]
    class is useful to apply a single transformation to a single data source, this
    class is helpful to apply multple transformations to a single data or multiple
    transformations to multiple data. This class is especially designed to be applied
    as part of `model` inside the `fit`, `predict` or `evaluate` methods. The
    `fit_transform` method should be applied before feeding the data to the
    algorithm and `inverse_transform` method should be called after algorithm has
    worked with data.

    Examples
    --------
        >>> import numpy as np
        >>> from ai4water.preprocessing.transformations import Transformations
        >>> x = np.arange(50).reshape(25, 2)
        >>> transformer = Transformations(['a', 'b'], config=['minmax', 'zscore'])
        >>> x_ = transformer.fit_transform(x)
        >>> _x = transformer.inverse_transform(x_)
        ...
        ... # Apply multiple transformations on multiple arrays which are passed as list
        >>> transformer = Transformations([['a', 'b'], ['a', 'b']],
        ...                              config=['minmax', 'zscore'])
        >>> x1 = np.arange(50).reshape(25, 2)
        >>> x2 = np.arange(50, 100).reshape(25, 2)
        >>> x1_transformed = transformer.fit_transform([x1, x2])
        >>> _x1 = transformer.inverse_transform(x1_transformed)

        We can also do more complicated stuff as following

        >>> transformer = Transformations({'x1': ['a', 'b'], 'x2': ['a', 'b']},
        ...        config={'x1': ['minmax', 'zscore'],
        ...                'x2': [{'method': 'log', 'features': ['a', 'b']},
        ...                       {'method': 'robust', 'features': ['a', 'b']}]
        ...                                      })
        >>> x1 = np.arange(20).reshape(10, 2)
        >>> x2 = np.arange(100, 120).reshape(10, 2)
        >>> x = {'x1': x1, 'x2': x2}
        >>> x_transformed = transformer.fit_transform(x)
        >>> _x = transformer.inverse_transform(x_transformed)

        In above example we apply `minmax` and `zscore` transformations on x1
        and `log` and `robust` transformations on x2 array
    """
    def __init__(
            self,
            feature_names: Union[list, dict],
            config: Union[str, list, dict] = None,
    ):
        """
        Arguments:
            feature_names:
                names of features in data
            config:
                Determines the type of transformation to be applied on data.
                It can be one of the following types

                - `string` when you want to apply single transformation

                >>> config='minmax'

                - `dict`: to pass additional arguments to the :py:class:`ai4water.preprocessing.Transformation`
                   class

                >>> config = {"method": 'log', 'treat_negatives': True, 'features': ['features']}

                - `list` when we want to apply multiple transformations

                >>> ['minmax', 'zscore']

                or

                >>> [{"method": 'log', 'treat_negatives': True, 'features': ['features']},
                >>> {'method': 'sqrt', 'treat_negatives': True}]


        """
        self.names = feature_names
        self.t_config = config
        self.without_fit = False

    def _fetch_transformation(self, data):
        config = self.t_config

        if isinstance(data, list):
            if isinstance(config, str):
                config = [config for _ in range(len(data))]
        elif isinstance(data, dict):
            if isinstance(config, str):
                config = {k:config for k in data.keys()}

        return config

    def _check_features(self):

        if self.is_numpy_:
            assert isinstance(self.names, list), f"""
            feature_names are of type {type(self.names)}"""

        elif self.is_list_:
            for idx, n in enumerate(self.names):
                assert isinstance(n, list), f"""
                feature_names for {idx} source is {type(n)}. It should be list"""

        elif self.is_dict_:
            assert isinstance(self.names, dict), f"""
            feature_names are of type {type(self.names)}"""
            for src_name, n in self.names.items():
                assert n.__class__.__name__ in ["ListWrapper", 'list']

        return

    def fit_transform(self, data:Union[np.ndarray, List, Dict]):
        """Transforms the data according the the `config`.

        Arguments:
            data:
                The data on which to apply transformations. It can be one of following

                - a (2d or 3d) numpy array
                - a list of numpy arrays
                - a dictionary of numpy arrays
        Returns:
            The transformed data which has same type and dimensions as the input data
        """
        setattr(self, 'is_numpy_', False)
        setattr(self, 'is_list_', False)
        setattr(self, 'is_dict_', False)
        setattr(self, 'scalers_', {})

        if self.t_config is None:  # if no transformation then just return the data as it is
            return data

        orignal_data_type = data.__class__.__name__

        if isinstance(data, np.ndarray):
            setattr(self, 'is_numpy_', True)
        elif isinstance(data, list):
            setattr(self, 'is_list_', True)
        elif isinstance(data, dict):
            setattr(self, 'is_dict_', True)
        else:
            raise ValueError(f"invalid data of type {data.__class__.__name__}")

        # first check that data matches config
        self._check_features()

        # then apply transformation
        data = self._fit_transform(data)

        # now pack it in original form
        assert data.__class__.__name__ == orignal_data_type, f"""
        type changed from {orignal_data_type} to {data.__class__.__name__}
        """

        #self._assert_same_dim(self, orignal_data, data)

        return data

    def _transform_2d(self, data, columns, transformation=None, key="5"):
        """performs transformation on single data 2D source"""
        # it is better to make a copy here because all the operations on data happen after this.
        data = data.copy()
        scalers = {}
        if transformation:

            if isinstance(transformation, dict):
                transformer = Transformation(**transformation)
                data = transformer.fit_transform(pd.DataFrame(data, columns=columns))
                scalers[key] = transformer.config()

            # we want to apply multiple transformations
            elif isinstance(transformation, list):
                for idx, trans in enumerate(transformation):

                    if isinstance(trans, str):
                        transformer = Transformation(method=trans)
                        data = transformer.fit_transform(pd.DataFrame(data, columns=columns))
                        scalers[f'{key}_{trans}_{idx}'] = transformer.config()

                    elif trans['method'] is not None:
                        transformer = Transformation(**trans)
                        data = transformer.fit_transform(pd.DataFrame(data, columns=columns))
                        scalers[f'{key}_{trans["method"]}_{idx}'] = transformer.config()
            else:
                assert isinstance(transformation, str)
                transformer = Transformation(method=transformation)
                data = transformer.fit_transform(pd.DataFrame(data, columns=columns))
                scalers[key] = transformer.config()

            data = data.values

        self.scalers_.update(scalers)

        return data

    def __fit_transform(self, data, feature_names, transformation=None, key="5"):
        """performs transformation on single data source
        In case of 3d array, the shape is supposed to be following
        (num_examples, time_steps, num_features)
        Therefore, each time_step is extracted and transfomred individually
        for example with time_steps of 2, two 2d arrays will be extracted and
        transformed individually
        (num_examples, 0,num_features), (num_examples, 1, num_features)
        """
        if data.ndim == 3:
            _data = np.full(data.shape, np.nan)
            for time_step in range(data.shape[1]):
                _data[:, time_step] = self._transform_2d(data[:, time_step],
                                                        feature_names,
                                                        transformation,
                                                        key=f"{key}_{time_step}")
        else:
            _data = self._transform_2d(data, feature_names, transformation, key=key)

        return _data

    def _fit_transform(self, data, key="5"):
        """performs transformation on every data source in data"""
        transformation = self._fetch_transformation(data)
        if self.is_numpy_:
            _data = self.__fit_transform(data, self.names, transformation, key)

        elif self.is_list_:
            _data = []
            for idx, array in enumerate(data):
                _data.append(self.__fit_transform(array,
                                                  self.names[idx],
                                                  transformation[idx],
                                                  key=f"{key}_{idx}")
                             )
        else:
            _data = {}
            for src_name, array in data.items():
                _data[src_name] = self.__fit_transform(array,
                                                       self.names[src_name],
                                                       transformation[src_name],
                                                       f"{key}_{src_name}")
        return _data

    def inverse_transform(self, data, postprocess=True):
        """inverse transforms data where data can be dictionary, list or numpy
        array.

        Arguments:
            data:
                the data which is to be inverse transformed. The output of
                `fit_transform` method.
            postprocess : bool

        Returns:
            The original data which was given to `fit_transform` method.
        """
        if not hasattr(self, 'scalers_'):
            raise ValueError(f"Transformations class has not been fitted yet")
        return self._inverse_transform(data, postprocess=postprocess)

    def inverse_transform_without_fit(self, data, postprocess=True)->np.ndarray:
        data = np.array(data)
        if data.ndim == 1:
            data = data.reshape(-1, 1)

        assert isinstance(self.names, list)
        assert data.shape[-1] == len(self.names)

        data = pd.DataFrame(data, columns=self.names)

        kwargs = {}
        if isinstance(self.t_config, str):
            kwargs['method'] = self.t_config
        elif isinstance(self.t_config, dict):
            kwargs = self.t_config
        elif isinstance(self.t_config, list):
            assert len(self.t_config) == 1
            t_config = self.t_config[0]
            if isinstance(t_config, str):
                kwargs['method'] = t_config
            elif isinstance(t_config, dict):
                kwargs = t_config
            else:
                raise ValueError(f"invalid type of t_config {t_config.__class__.__name__}")
        else:
            raise ValueError(f"invalid type of t_config {self.t_config.__class__.__name__}")

        transformer = Transformation(**kwargs)
        transformed_data = transformer.inverse_transform(data=data, postprocess=postprocess)

        return transformed_data.values

    def _inverse_transform(self, data, key="5", postprocess=True):

        transformation = self._fetch_transformation(data)

        if self.is_numpy_:
            data = self.__inverse_transform(data,
                                            self.names,
                                            transformation,
                                            key,
                                            postprocess=postprocess)

        elif self.is_list_:
            assert isinstance(data, list)
            _data = []
            for idx, src in enumerate(data):
                __data = self.__inverse_transform(src,
                                                 self.names[idx],
                                                 transformation[idx],
                                                 f'{key}_{idx}',
                                                  postprocess=postprocess)
                _data.append(__data)
            data = _data

        elif self.is_dict_:
            assert isinstance(data, dict)
            _data = {}
            for src_name, src in data.items():
                _data[src_name] = self.__inverse_transform(src,
                                                          self.names[src_name],
                                                          transformation[src_name],
                                                          f'{key}_{src_name}',
                                                           postprocess=postprocess)
            data = _data

        return data

    def __inverse_transform(self,
                            data,
                            feature_names,
                            transformation, key="5",
                            postprocess=True):
        """inverse transforms one data source which may 2d or 3d nd array"""
        if data.ndim == 3:
            _data = np.full(data.shape, np.nan)
            for time_step in range(data.shape[1]):
                _data[:, time_step] = self._inverse_transform_2d(
                    data[:, time_step],
                    columns=feature_names,
                    transformation=transformation,
                    key=f"{key}_{time_step}",
                    postprocess=postprocess)
        else:
            _data = self._inverse_transform_2d(data,
                                               feature_names,
                                               key,
                                               transformation,
                                               postprocess=postprocess)

        return _data

    def _inverse_transform_2d(self,
                              data,
                              columns,
                              key,
                              transformation,
                              postprocess=True)->np.ndarray:
        """inverse transforms one 2d array"""
        data = pd.DataFrame(data.copy(), columns=columns)

        if transformation is not None:
            if isinstance(transformation, str):

                if key not in self.scalers_:
                    raise ValueError(f"""
                    key `{key}` for inverse transformation not found. Available keys are {list(self.scalers_.keys())}""")

                scaler = self.scalers_[key]
                scaler, shape = scaler, scaler['shape']
                original_shape = data.shape

                transformer = Transformation.from_config(scaler)
                transformed_data = transformer.inverse_transform(data, postprocess=postprocess)
                data = transformed_data

            elif isinstance(transformation, list):
                # idx and trans both in reverse form
                for idx, trans in reversed(list(enumerate(transformation))):
                    if isinstance(trans, str):
                        scaler = self.scalers_[f'{key}_{trans}_{idx}']
                        scaler, shape = scaler, scaler['shape']
                        transformer = Transformation.from_config(scaler)
                        data = transformer.inverse_transform(data=data, postprocess=postprocess)

                    elif trans['method'] is not None:
                        features = trans.get('features', columns)
                        # if any of the feature in data was transformed
                        if any([True if f in data else False for f in features]):
                            orig_cols = data.columns  # copy teh columns in the original df
                            scaler = self.scalers_[f'{key}_{trans["method"]}_{idx}']
                            scaler, shape = scaler, scaler['shape']
                            data, dummy_features = conform_shape(data, shape, features)  # get data to transform

                            transformer = Transformation.from_config(scaler)
                            transformed_data = transformer.inverse_transform(data=data,
                                                                             postprocess=postprocess)
                            data = transformed_data[orig_cols]  # remove the dummy data

            elif isinstance(transformation, dict):

                features = transformation.get('features', columns)
                if any([True if f in data else False for f in features]):
                    orig_cols = data.columns
                    scaler = self.scalers_[key]
                    scaler, shape = scaler, scaler['shape']
                    data, dummy_features = conform_shape(data, shape, features=features)

                    transformer = Transformation.from_config(scaler)
                    transformed_data = transformer.inverse_transform(data=data, postprocess=postprocess)
                    data = transformed_data[orig_cols]  # remove the dummy data

        if data.__class__.__name__ == "DataFrame":
            data = data.values  # there is no need to return DataFrame

        return data

    def config(self)->dict:
        """returns a python dictionary which can be used to construct this class
        in fitted form i.e as if the fit_transform method has already been applied.
        Returns:
            a dictionary from which `Transformations` class can be constructed
        """
        return {
            'scalers_': jsonize(self.scalers_),
            "feature_names": self.names,
            "config": self.t_config,
            "is_numpy_": self.is_numpy_,
            "is_dict_": self.is_dict_,
            "is_list_": self.is_list_,
        }

    @classmethod
    def from_config(cls, config:dict)->"Transformations":
        """constructs the Transformations class which may has already been fitted.
        """
        config = deepcopy_dict_without_clone(config)

        transformer = cls(config.pop('feature_names'), config.pop('config'))

        for attr_name, attr_val in config.items():
            setattr(cls, attr_name, attr_val)

        return transformer


def conform_shape(data, shape, features=None):
    # if the difference is of only 1 dim, we resolve it
    if data.ndim > len(shape):
        data = np.squeeze(data, axis=-1)
    elif data.ndim < len(shape):
        data = np.expand_dims(data, axis=-1)

    assert data.ndim == len(shape), f"""original data had {len(shape)} wihle the 
    new data has {data.ndim} dimensions"""

    # how manu dummy features we have to add to match the shape
    dummy_features = shape[-1] - data.shape[-1]

    if data.__class__.__name__ in ['DataFrame', 'Series']:
        # we know what features must be in data, so put them in data one by one
        # if they do not exist in data already
        if features:
            for f in features:
                if f not in data:
                    data[f] = np.random.random(len(data))
        # identify how many features to be added by shape information
        elif dummy_features > 0:
            dummy_data = pd.DataFrame(np.random.random((len(data), dummy_features)))
            data = pd.concat([dummy_data, data], axis=1)
    else:
        dummy_data = np.random.random((len(data), dummy_features))
        data = np.concatenate([dummy_data, data], axis=1)

    return data, dummy_features
